import numpy as np
from .mlfm import BaseMLFM, Dimensions
from scipy.stats import norm
from scipy.optimize import minimize
from scipy.linalg import cho_solve

def _unpack_vector(x, shape):
    res = []
    ntot = 0
    for n in shape:
        res.append(x[ntot:ntot+n])
        ntot += n
    return res

def _get_gp_theta_shape(gp_list):
    """ Handler function for getting the shape of the
    free hyper parameters for the a list of Gaussian Processes objects
    """
    return [gp.kernel.theta.size for gp in gp_list]

def _var_mixer(free_vars, free_vars_shape, fixed_vars, is_fixed_vars):
    """
    Utility function to mix the free vars and the fixed vars
    """
    free_vars = _unpack_vector(free_vars, free_vars_shape)
    if is_fixed_vars is None:
        return free_vars
    else:
        full_vars = []
        ifree, ifixed = (0, 0)
        for b in is_fixed_vars:
            if b:
                full_vars.append(fixed_vars[ifixed])
                ifixed += 1
            else:
                full_vars.append(free_vars[ifree])
                ifree += 1
        return full_vars    


# Utility functions for padding the time vector
def augment_times(t, h):
    """
    Pads the vector t so that the maximum spacing is h.

    Parameters
    ----------

    t : one dimensional array like

    h : float
    """
    inds = [0]
    res = [t[0]]
    for ta, tb in zip(t[:-1], t[1:]):
        N = np.ceil((tb - ta) / h + 1)
        N = int(N)
        _tt = np.linspace(ta, tb, N)
        res = np.concatenate((res, _tt[1:]))
        inds.append(inds[-1] + N - 1)
    return res, np.array(inds)

def handle_time_inds(times, h=None):
    """
    Takes a list of time vectors and returns the unique, potentially augmented,
    vector 
    """
    # get size of each vector
    t_sizes = [len(t) for t in times]

    # concatenate to single time vector
    tt = np.concatenate(times)

    # get the distinct times, and the indices
    tt_uni, inv_ind = np.unique(tt, return_inverse=True)

    # split inv_ind up
    ind_ti = _unpack_vector(inv_ind, t_sizes)

    if h is None:
        return tt_uni, ind_ti

    elif isinstance(h, float) and h > 0:
        # augment the time vector so that diff is at most h
        ttc, inds_c = augment_times(tt_uni, h)
        data_inds = [inds_c[ind_i] for ind_i in ind_ti]
        return ttc, data_inds

    else:
        raise ValueError("h should be a float > 0")

def Mpower_action_wgrad(M, dM, x, power, dMdb=None):
    """
    Returns the result of M^{power}x and its gradient. 
    """
    res = x
    grad = [np.zeros(x.size).reshape(x.shape)]*len(dM)

    if dMdb is not None:
        grad_b = [np.zeros(x.size).reshape(x.shape)]*len(dMdb)
    
    for i in range(power):
        grad = [dM_i.dot(res) + M.dot(grad_i)
                for dM_i, grad_i in zip(dM, grad)]
        if dMdb is not None:

            grad_b = [dM_i.dot(res) + M.dot(grad_i)
                      for dM_i, grad_i in zip(dMdb, grad_b)]
            
        res = M.dot(res)

    if dMdb is None:
        return res, np.array(grad)
    else:
        return res, grad, grad_b

class EMFitMixin:
    """
    Handles MAP fitting of the MLFM-SA-Mixture model
    """

    def _em_fit(self, init, free_vars_shape, fixed_vars, is_fixed_vars, priors,
                max_nt=20, **kwargs):
        free_vars = init.copy()
        g, beta, mu_ivp = _var_mixer(init, free_vars_shape, fixed_vars, is_fixed_vars)
        beta = beta.reshape(self.dim.R+1, self.dim.D)
        mu_ivp = mu_ivp.reshape((len(self.x_train_), len(self._ifix), self.dim.K))

        pi = [np.ones(len(self._ifix)) / len(self._ifix) ]*len(self.x_train_)
        resp = self._get_responsibilities(pi, g, beta, mu_ivp, 1000, self._ifix)

        gcur = g.copy()
        for nt in range(max_nt):

            if nt < 5:
                is_nt_early = True
            else:
                is_nt_early = False

            import time
            
            # M-step
            t0 = time.time()
            free_vars = self._M_step(free_vars, resp, 2000,
                                     free_vars_shape, fixed_vars, is_fixed_vars,
                                     priors,
                                     is_nt_early=is_nt_early,
                                     **kwargs)
            t1 = time.time()
            print("Time taken for M step: ", t1-t0)
            pi = self._update_pi(resp)

            # E-Step
            # update responsibilites
            t0 = time.time()
            g, beta, mu_ivp = _var_mixer(free_vars, free_vars_shape, fixed_vars, is_fixed_vars)
            beta = beta.reshape(self.dim.R+1, self.dim.D)
            mu_ivp = mu_ivp.reshape((len(self.x_train_), len(self._ifix), self.dim.K))
            dg = np.linalg.norm(g - gcur)
            if dg < 1e-4:
                break
            else:
                # keep going
                gcur = g
                resp = self._get_responsibilities(pi, g, beta, mu_ivp, 2000, self._ifix)
            t1 = time.time()                
            print("Time take for E step ", t1-t0)


        return g, beta, pi, mu_ivp

    def _M_step(self, free_vars,
                responsibilites, alpha_prec,
                free_vars_shape, fixed_vars, is_fixed_vars, priors,
                is_nt_early=False, **kwargs):
        """Optimize wrt to free variables
        """
        logpsi = np.concatenate([gp.kernel.theta for gp in self.latentforces])


        # Has to be a nicer way of doing this using slices
        inds = [np.concatenate([self.data_inds[q] + k*self.dim.N
                                for k in range(self.dim.K)])
                for q in range(len(self.y_train_))]

        def objfunc(arg, free_vars_shape, fixed_vars):
            # unpack the vector
            g, vbeta, mu_ivp = _var_mixer(arg, free_vars_shape, fixed_vars, is_fixed_vars)
            mu_ivp = mu_ivp.reshape((len(self.x_train_), len(self._ifix), self.dim.K))
            beta = vbeta.reshape(self.dim.R+1, self.dim.D)
            loglik = 0.
            loglik_g_grad = 0.
            loglik_b_grad = 0.

            try:
                # sum over the mixture components
                for m, ifx in enumerate(self._ifix):
                    ll, ll_g_grad, ll_b_grad = \
                        self.log_likelihood_k(g, beta, mu_ivp[:, m, :], alpha_prec,
                                              responsibilites, m, ifx, inds,
                                              True, True)
                    loglik += ll
                    loglik_g_grad += ll_g_grad
                    loglik_b_grad += ll_b_grad

                logprior, logprior_g_grad, _ = self._prior_logpdf(g, logpsi, True)

                grad = [(loglik_g_grad + logprior_g_grad),
                        loglik_b_grad]

                # contributions from priors
                for vn, indx in zip(['beta',], [(1, vbeta),]):
                    try:
                        prior_logpdf = priors[vn]
                        ind, x = indx  # unpack the value of var name
                        vn_lp, vn_lp_grad = prior_logpdf(x, True)
                        logprior += vn_lp
                        grad[ind] += vn_lp_grad

                    except KeyError:
                        pass

                grad = np.concatenate([item for item, b in zip(grad, is_fixed_vars)
                                       if not b])
                return -(loglik + logprior), -grad
    
            except np.linalg.LinAlgError:
                return np.inf, np.zeros(arg.size)

        try:
            optimopts = kwargs['optimopts'].copy()
        except KeyError:
            optimopts = {'disp': True}

        if is_nt_early:
            # reduce the number of steps to speed up search
            # in early states
            optimopts['maxiter'] = 50
            optimopts['gtol'] = 1e-3

        res = minimize(objfunc, free_vars,
                       jac=True,
                       args=(free_vars_shape, fixed_vars),
                       method='BFGS',
                       options = optimopts)

        self._optimres = res  # store a copy of the results

        return res.x


    def _get_responsibilities(self, pi, g, beta, mu_ivp, alpha_prec, ifix):
        """ Gets the posterior responsibilities for each comp. of the mixture
        
        Returns
        -------
            r : list, [r1, ..., rm]
            Returns the responsibilities for each of the m outputs.
        """
        probs = [[]]*len(self.N_data)  # store the log probabilities for each output

        # sum over the number of components
        for i, ifx in enumerate(ifix):

            # Get the transition matrix
            K = self._K(g, beta, ifx)
            # Raise it to a power
            Lop = np.linalg.matrix_power(K, self.order)

            # the means for each output [NK, N_outputs]
            means = np.kron(mu_ivp[:, i, :],
                            np.ones(self.dim.N)).T
            means = Lop.dot(means)

            for q, ym in enumerate(self.y_train_):
                # subsample means using data inds
                m_i_q = means[:, q].reshape(self.dim.K, self.dim.N).T

                m_i_q = m_i_q[self.data_inds[q], :]
                lp_i_q = norm.logpdf(
                    ym.reshape(self.dim.K, self.N_data[q]).T,
                    loc=m_i_q,
                    scale=1/np.sqrt(alpha_prec))
                # sum over the dimesnions
                lp_i_q = lp_i_q.sum(-1)

                if probs[q] == []:
                    probs[q] = lp_i_q[:, None]
                else:
                    probs[q] = np.column_stack((probs[q], lp_i_q))

        # weight by mixture component probs.
        probs = [lp - pi_m[None, :]
                 for lp, pi_m in zip(probs, pi)]
        # subtract the maximum for exponential normalize
        probs = [p - p.max(axis=-1)[:, None] for p in probs]
        probs = [np.exp(p) / np.exp(p).sum(-1)[:, None] for p in probs]
        return probs

    def _update_pi(self, responsibilities):
        """Update the mixture probabilities
        """
        pi_mean = [np.mean(rm, axis=0) for rm in responsibilities]
        pi_mean = sum(pi_mean) / len(pi_mean)
        return [pi_mean] * len(responsibilities)


def kron_A_enT(A, n, N):
    """
    Fast calculation of the kronecker product
    of a matrix A with the size N basis vector
    transposed.
    """
    res = np.zeros((A.shape[0], A.shape[1]*N))
    res[:, n::N] = A
    return res

class BaseMLFMSA(BaseMLFM):

    is_tt_aug = False
    def __init__(self, *args, order=1, **kwargs):
        super(BaseMLFMSA, self).__init__(*args, **kwargs)

        if isinstance(order, int) and order > 0:
            self.order = order

        else:
            raise ValueError("The approximation error must be a "
                             " postive integer.")

    def _K(self, g, beta, ifix, eval_gradient=False, eval_b_gradient=False):
        """
        Return the discretisation of the integral opearator in Picard iteration.
        """
        W = self._get_weight_matrix(self.ttc, ifix)
        g = g.reshape((self.dim.R, self.dim.N)).T
        g = np.column_stack((np.ones(self.dim.N), g))
        
        # struct matrices A_r = sum_r g_rn * sum_d brd * Ld
        struct_mats = np.array([sum(brd*Ld
                                    for brd, Ld in zip(br, self.basis_mats))
                                for br in beta])
        #At = struct_mats[None, ...] * g[..., None, None]
        #At = At.sum(1)  # sum over R
        
        if eval_gradient:

            #I = np.eye(self.dim.K)
            
            K, K_grad = (0, [])
            for n in range(self.dim.N):
                #Anw = np.kron(At[n, ...], W[:, n][:, None])
                #K += kron_A_enT(Anw, n, self.dim.N)
                for r in range(self.dim.R):
                    Ar_x_wn = np.kron(struct_mats[r+1], W[:, n][:, None])
                    K_grad.append(kron_A_enT(Ar_x_wn, n, self.dim.N))

                # equivalent (should be a much faster way of doing this... its all zeros)
                #Ien = np.kron(I, np.eye(1, self.dim.N, n).T)
                #Ien = np.zeros((self.dim.N*self.dim.K, self.dim.K))
                #Ien[np.arange(n, self.dim.N*self.dim.K, self.dim.N),
                #np.arange(self.dim.K)] = 1.
                #
                #K += kron_A_enT(Ien, ifix, self.dim.N)

            if eval_b_gradient:
                # This is lazy
                dB = self._dKdB(g, beta, ifix)
                K = sum(brd * item for brd, item in zip(beta.ravel(), dB))
                for k in range(self.dim.K):
                    K[k*self.dim.N:(k+1)*self.dim.N, ifix+k*self.dim.N] += 1.

                return K, K_grad, dB

            return K, K_grad
        else:
            At = struct_mats[None, ...] * g[..., None, None]
            At = At.sum(1)  # sum over R
            eifx = np.eye(1, self.dim.N, ifix)
            K = 0.
            for n in range(self.dim.N):
                Anw = np.kron(At[n, ...], W[:, n][:, None])
                K += kron_A_enT(Anw, n, self.dim.N)
            
            I = np.eye(self.dim.K)
        
            K += sum(np.kron(I, np.kron(np.eye(1, self.dim.N, i).T, eifx))
                     for i in range(self.dim.N))
            return K

    def _dKdB(self, g, beta, ifix):
        """
        Rewriting of this function to reduce redundant computations
        """
        W = self._get_weight_matrix(self.ttc, ifix)
        

        # for computational efficiency we need to do
        # any scalar multiplications *before* any
        # kronecker products
        GW = np.stack((gn[:, None] * W[:, n][None, :]
                       for n, gn in enumerate(g)))

        dKdB = []
        for r in range(self.dim.R+1):
            for d in range(self.dim.D):
                dKdB.append(np.kron(self.basis_mats[d],
                                    GW[:, r, :].T))

        return dKdB

    def _get_weight_matrix(self, tt, i, store=True):
        """
        Weight matrix for the trapezoidal rule.
        """
        try:
            return self._weights[i]
        except:

            def _get_weights(tt):
                tt = np.asarray(tt)
                W = np.zeros((tt.size, tt.size))
                h = np.diff(tt)
                for i in range(len(tt)):
                    W[i, :i] += .5 * h[:i]
                    W[i, 1:i+1] += .5 * h[:i]
                return W

            ttb = tt[:i+1]
            ttf = tt[i:]

            Wb = _get_weights([t for t in reversed(ttb)])
            Wb = np.flip(Wb, (0, 1))
            Wf = _get_weights(ttf)

            W = np.zeros((len(tt), len(tt)))
            W[:i+1, :i+1] += Wb
            W[i:, i:] += Wf

            if store:
                if hasattr(self, '_weights'):
                    self._weights[i] = W
                else:
                    self._weights = {i: W}

            return W


class FitMixin:
    """
    Adds utility functions for model fitting.
    """
    def fit(self, experiments, ifix, **kwargs):
        """
        Parameters
        ----------
        experiments : A list-like of tuples [(ti, Yi)]
            Each entry (ti, Yi) represents a set of time points
            and observations Yi from the model

        """
        method = 'EM'

        # make sure the model is reader for fitting by calling _setup

        times = []
        Y = []
        for E in experiments:
            times.append(E[0])
            Y.append(E[1])
        
        self._setup(times, ifix, **kwargs)
        self.y_train_ = [y.T.ravel() for y in Y]


        # parse kwargs to see if any args are being fixed
        is_fixed_vars = self._fit_kwarg_parser(**kwargs)

        # get the initial values
        init, free_vars_shape, fixed_vars = \
              self._fit_init(is_fixed_vars, **kwargs)

        # parse kwargs for priors
        priors = {}
        for vn in ['beta']:
            key = '_'.join((vn, 'prior'))
            try:
                priors[vn] = kwargs[key].logpdf
            except KeyError:
                pass

        if method == 'EM':
            return self._em_fit(init, free_vars_shape, fixed_vars, is_fixed_vars, priors, **kwargs)
        
    def _setup(self, times, ifix, **kwargs):
        """Prepares the model for fitting."""
        if not hasattr(self, 'ttc'):
            self._setup_times(times, **kwargs)

        self._setup_ifix(ifix)

    def _setup_ifix(self, ifix):
        # Optionally automate this in future?
        self._ifix = ifix

    def _setup_times(self, tt,
                     h=None, tt_aug=None,
                     multi_output=False):
        """ Handles storing of training times and
        augmenting the time vector
        """
        # numpy arrayify time vectors
        tt = [np.asarray(item) for item in tt]
        
        if True:# h is not None:
            # h is checked first
            ttc, data_inds = handle_time_inds(tt, h)
            self.x_train_ = tt
            self.ttc = ttc
            self.data_inds = data_inds
            self.is_tt_aug = True

        elif tt_aug is not None:
            # User has supplied the time points at which they want
            # to estimate the latent forces

            if multi_output:
                # add tt_aug to the times list
                tt = [tt_aug] + [t for t in tt]
                ttc, data_inds = handle_time_inds(tt, h)
                self.x_train_ = tt[1:]
                self.ttc = ttc
                self.data_inds = data_inds[1:]
                self.is_tt_aug = True
        else:
            # neither h or tt_aug is supplied
            self.x_train_ = tt
            self.ttc = ttc
            self.is_tt_aug = None

        # store the dimension variables
        _, K, R, D = self.dim
        self.dim = Dimensions(self.ttc.size, K, R, D)
        self.N_data = tuple(item.size for item in self.x_train_)

    def _fit_init(self, is_fixed_vars, **kwargs):
        """ Handles model initialisation.
        """
        init_strategies = {
            'g': lambda : np.zeros(self.dim.N*self.dim.R),
            'beta': lambda : np.row_stack((np.zeros(self.dim.D),
                                           np.eye(self.dim.R, self.dim.D))).ravel(),
            'mu_ivp': lambda : np.zeros((len(self.x_train_), len(self._ifix), self.dim.K)),
            }

        var_names = ['g', 'beta', 'mu_ivp']
        full_init = [kwargs.pop("".join((vn, '0')),
                                init_strategies[vn]()).ravel()
                     for vn in var_names]
        full_init_shape = [item.size for item in full_init]

        free_vars, fixed_vars = ([], [])
        for item, boolean in zip(full_init, is_fixed_vars):
            if boolean:
                fixed_vars.append(item)
            else:
                free_vars.append(item)
        free_vars_shape = [item.size for item in free_vars]

        return np.concatenate(free_vars), free_vars_shape, fixed_vars

    def _fit_kwarg_parser(self, **kwargs):
        var_names = ['g', 'beta', 'mu_ivp']
        is_fixed_vars = [kwargs.pop("".join((vn, "_is_fixed")), False)
                         for vn in var_names]
        return is_fixed_vars
            
class MLFMSAMix(EMFitMixin, FitMixin, BaseMLFMSA):
                
    """
    Extends the MLFMSA model to use mixtures
    """

    def log_likelihood_k(self,
                         g, beta,
                         mu_ivp, alpha_prec,
                         responsibilites,
                         m, ifx, inds,
                         eval_g_grad=False, eval_b_grad=False):
        """ Returns the log-likelihood of a component of the mixture
        """

        Km, Km_g_grad, Km_b_grad = self._K(g, beta, ifx,
                                           eval_g_grad, eval_g_grad)

        means = np.kron(mu_ivp,
                        np.ones(self.dim.N)).T

        # raise K to self.order and get the gradient of its action on
        # means
        L, Ldg, Ldb = Mpower_action_wgrad(Km, Km_g_grad, means, self.order, Km_b_grad)
        
        log_lik, log_lik_g_grad, log_lik_b_grad = (0, 0, 0)

        resp = [np.row_stack([r_q]*self.dim.K)
                for r_q in responsibilites]
        
        # sum over the number of outputs
        for q, y in enumerate(self.y_train_):
            inds_q = inds[q]
            Lim_q = L[inds_q, q]

            eta = Lim_q - y
            r_q_i = resp[q][:, m]        # responsibilities for the qth replicate
                                         # of the ith mixture component
            Reta = eta * r_q_i           # weight eta by the responsibilites

            # update the log-likelihood and gradient
            log_lik += -0.5 * (eta * Reta).sum() * alpha_prec
            log_lik_g_grad -= alpha_prec * np.array(
                [np.sum(Reta * dL[inds_q, q]) for dL in Ldg])
            log_lik_b_grad -= alpha_prec * np.array(
                [np.sum(Reta * dL[inds_q, q]) for dL in Ldb])

        return log_lik, log_lik_g_grad, log_lik_b_grad

    def log_likelihood(self, g, beta,
                       mu_ivp, alpha_prec, pi,
                       ifix):

        # Has to be a nicer way of doing this using slices
        inds = [np.concatenate([self.data_inds[q] + k*self.dim.N
                                for k in range(self.dim.K)])
                for q in range(len(self.y_train_))]
        
        logps = []
        for m, ifx in enumerate(ifix):

            Km = self._K(g, beta, ifx)
            means = np.kron(mu_ivp[:, m, :], np.ones(self.dim.N)).T

            L = np.linalg.matrix_power(Km, self.order)
            L = L.dot(means)

            # sum over the experiments
            logp_m = 0.

            for q, y in enumerate(self.y_train_):
                inds_q = inds[q]
                Lim_q = L[inds_q, q]
                eta = Lim_q - y

                logp_m += norm.logpdf(eta,
                                      scale=1/np.sqrt(alpha_prec)).sum()
            logps.append(logp_m + np.log(pi[m]))

        logps = np.array(logps)
        a = logps.max()

        res = a + np.log(np.exp(logps - a).sum())
        return res

    def _laplace(self, g, beta, mu_ivp, alpha_prec, pi):

        logpsi = np.concatenate([gp.kernel.theta
                                 for gp in self.latentforces])

        def foo(g):
            ll = self.log_likelihood(g, beta, mu_ivp, alpha_prec, pi, self._ifix)
            lp = self._prior_logpdf(g, logpsi)
            return -(ll + lp)

        res = minimize(foo, g, options={'maxiter' : 20})
        C = res.hess_inv
        return res.x, C
        
    def _prior_logpdf(self, g, logpsi, eval_gradient=False, **kwargs):
        veclogpsi = logpsi.copy()  # need to keep a copy of the vectorised var.
        logpsi_shape = _get_gp_theta_shape(self.latentforces)
        logpsi = _unpack_vector(logpsi, logpsi_shape)

        # reshape g
        g = g.reshape(self.dim.R, self.dim.N)
        ll = 0.

        if eval_gradient: ll_g_grad, ll_logpsi_grad = ([], [])
        
        for r in range(self.dim.R):
            gp = self.latentforces[r]
            kern = gp.kernel.clone_with_theta(logpsi[r])

            if eval_gradient:
                K, K_gradient = kern(self.ttc[:, None],
                                     eval_gradient=True)
            else:
                K = kern(self.ttc[:, None])


            K[np.diag_indices_from(K)] += gp.alpha
            L = np.linalg.cholesky(K)

            alpha = cho_solve((L, True), g[r, :])

            tmp = -.5 * g[r, :].dot(alpha)
            tmp -= np.log(np.diag(L)).sum()

            ll += -.5 * g[r, :].dot(alpha) - \
                  np.log(np.diag(L)).sum() - \
                  L.shape[0] / 2 * np.log(2 * np.pi)

            if eval_gradient:
                ll_g_grad.append(-alpha)
                
                tmp = np.outer(alpha, alpha)
                tmp -= cho_solve((L, True), np.eye(K.shape[0]))
                ll_logpsi_grad.append(.5 * np.einsum("ij,ijk", tmp, K_gradient))

        # concatenate things together
        if eval_gradient:
            ll_logpsi_grad = np.concatenate(ll_logpsi_grad)

        ## Inspect kwargs for any additionally passed priors
        try:
            logpsi_prior = kwargs['logpsi_prior']
            if eval_gradient:
                lp_psi, lp_psi_grad = logpsi_prior.logpdf(veclogpsi, eval_gradient=True)
                ll_logpsi_grad += lp_psi_grad
            else:
                lp_psi = logpsi_prior(logpsi, eval_gradient=False)
            ll += lp_psi  # add the contribution from psi prior

        except KeyError:
            # Ignore it if no logpsi prior passed.
            pass
            

        if eval_gradient:
            return ll, \
                   np.concatenate(ll_g_grad), \
                   ll_logpsi_grad

        return ll



import numpy as np
import scipy.sparse as sparse

def _unpack_vector(x, xk_shape):
    """
    Unpacks a flat vector into the len(xk_shape) subvectors
    of size xk_shape[i], i=1,...,len(xk_shape)
    """
    ntot = 0
    res = []
    for nk in xk_shape:
        res.append(x[ntot:ntot + nk])
        ntot += nk
    return res

def T_xrav_tovecx(N, M):
    """ Returns the NM x NM sparse matrix mapping X.ravel() to vec(X)
    """
    T = []
    for n in range(N*M):
        # image of a basis vector
        c = n - (n // M)*M   # col.
        r = n // M           # row
        # transpose
        c, r = (r, c)

        Ten = sparse.lil_matrix((N*M, 1))
        Ten[r*N + c] = 1.
        T.append(Ten)

    return sparse.hstack(T)

def _get_gp_theta_shape(gp_list):
    """ Handler function for getting the shape of the
    free hyper parameters for the a list of Gaussian Processes objects
    """
    return [gp.kernel.theta.size for gp in gp_list]



import numpy as np
import scipy.sparse as sparse
from sklearn.gaussian_process.kernels import ConstantKernel, RBF
from sklearn.gaussian_process import GaussianProcessRegressor
from scipy.linalg import block_diag, cho_solve, cholesky
from scipy.optimize import minimize
from .mlfm import BaseMLFM, Dimensions
from collections import namedtuple
from . import util

MLFMSuccApproxFitResults = namedtuple('MLFMSuccApproxFitResults',
                                      'g, beta, logpsi, logphi, loggamma, logtau, logalpha, \
                                      optimres')


class BaseMLFMSuccApprox(BaseMLFM):

    is_tt_aug = False

    def __init__(self, *args, order=1, **kwargs):
        super(BaseMLFMSuccApprox, self).__init__(*args, **kwargs)

        if isinstance(order, int) and order > 0:
            self.order = order
        else:
            raise ValueError("The approximation order must be a"
                             " positive integer")

    def _setup_latentstates(self, kernels=None, **kwargs):
        """
        Create the initial GP approximation
        """
        if kernels is None:
            ls_kernels = [ConstantKernel()*RBF() for k in range(self.dim.K)]
            self.latentstates = [GaussianProcessRegressor(kern)
                                 for kern in ls_kernels]
        else:
            self.latentstates = [GaussianProcessRegressor(kern) for kern in ls_kernels]

    def _setup_times(self, tt, tt_aug=None, ifix=0, **kwargs):
        if tt_aug is not None:
            self.ttc = tt_aug.copy()
            self._tt_data = tt.copy()
            self.data_inds = kwargs['data_inds']
            self._Ndata = len(self.data_inds)
            self.is_tt_aug = True
        else:
            self.ttc = tt
        self.ifix = ifix
        self.dim = Dimensions(self.ttc.size, self.dim.K, self.dim.R, self.dim.D)
        self._weight_matrix = weight_matrix(self.ttc, ifix)

    def _setup(self, times, **kwargs):
        """Prepares the model for fitting
        """
        if not hasattr(self, 'ttc'):
            self._setup_times(times, **kwargs)

        if not hasattr(self, 'latentstates'):
            # no latent states supplied so use default setting
            self._setup_latentstates(**kwargs)

    def _K(self, g, beta, ifix=None):
        """ Returns the discretised integral operator
        """
        if ifix is None:
            ifix = self.ifix
    
        g = g.reshape(self.dim.R, self.dim.N).T        # set g as g_{nr}, n=1..N, r=1..R
        g = np.column_stack((np.ones(self.dim.N), g))  # append a col. of 1s to g

        # matrices A[r] = sum(b_rd * Ld
        struct_mats = np.array([sum(brd*Ld for brd, Ld in zip(br, self.basis_mats))
                                for br in beta])
        # Add an axis for n=0,...,N-1
        At = struct_mats[None, :, :, :] * g[..., None, None]
        At = At.sum(1)

        W = weight_matrix(self.ttc, ifix) #self._weight_matrix
        K = sum(sparse.kron(sparse.kron(W[:, i][:, None],
                                        sparse.eye(1, self.dim.N, i)),
                            At[i, ...])
                for i in range(self.dim.N))
        I = sparse.eye(self.dim.K)

        # because x[ifix] is fixed by the integral transform
        # we need to add a column of identity matrices to K
        # - could do this by assignment
        eifix = sparse.eye(1, self.dim.N, ifix)
        K += sum(sparse.kron(sparse.kron(sparse.eye(1, self.dim.N, i).transpose(),
                                         eifix),
                             I)
                 for i in range(self.dim.N))
        return K

    def _K2(self, g, beta, ifix=None):
        """ K acting on vec(X) rather than ravel(X)
        """
        K = self._K(g, beta, ifix)
        T = util.T_xrav_tovecx(self.dim.N, self.dim.K)
        Tt = T.transpose()
        return T.dot(K.dot(Tt))

    def get_weight_matrix(self, tt, i):
        return weight_matrix(self.ttc, i)

    def _K3(self, g, beta, ifix, printit=False):
        W = weight_matrix(self.ttc, ifix)

        g = g.reshape(self.dim.R, self.dim.N).T
        g = np.column_stack((np.ones(self.dim.N), g))

        # matrices A[r] = sum(b_rd * Ld
        struct_mats = np.array([sum(brd*Ld for brd, Ld in zip(br, self.basis_mats))
                                for br in beta])
        # Add an axis for n=0,...,N-1
        At = struct_mats[None, :, :, :] * g[..., None, None]
        At = At.sum(1)

        eifix = np.eye(1, self.dim.N, ifix)
        K = sum(np.kron(At[i,...],
                        np.kron(W[:, i][:, None], np.eye(1, self.dim.N, i)))
                        for i in range(self.dim.N))

        I = np.eye(self.dim.K)
        eifix = np.eye(1, self.dim.N, ifix)
        K += sum(np.kron(I,
                         np.kron(np.eye(1, self.dim.N, i).T, eifix))
                 for i in range(self.dim.N))
        return K

    def sparse_vecK_aff_rep(self, beta):
        """Returns the vectorisation of the discretised integral
        operator
        """
        V0 = self._K(np.zeros(self.dim.N*self.dim.R), beta)  # lazy way of finding K[0]
        W = self._weight_matrix

        # A[r] = sum_{brd} * Ld
        struct_mats = np.array([sum(brd*Lrd for brd, Lrd in zip(br, self.basis_mats))
                                for br in beta])

        # columns of dK.dot(g) + vec(V0) = vec(K[g, beta])
        NK = self.dim.N*self.dim.K
        K = self.dim.K

        cols = []
        for r in range(self.dim.R):
            for n in range(self.dim.N):
                v1 = sparse.csr_matrix((n*K*NK, 1))          # Padding of zeros either
                v3 = sparse.csr_matrix(((NK-(n+1)*K)*NK,1))  # side of the vector
                v2 = np.kron(W[:, n][:, None], struct_mats[r+1]).T.ravel()[:, None]
                cols.append(sparse.vstack((v1, v2, v3)))

        V = sparse.hstack(cols)
        v0 = sparse.coo_matrix(V0.todense().T.ravel()).transpose()  # ugly step

        # convert V, v0 to sparse csr (quick for .dot( some_dense_arr )
        return V.tocsr(), v0.tocsr()

    @property
    def Ndata(self):
        if self.is_tt_aug:
            return len(self.data_inds)
        else:
            return self.dim.N


class MLFMSAFitMixin:
    """
    Utility mixin for initialising fit functions
    """
    def _fit_kwarg_parser(self, **kwargs):
        """
        Parses kwargs to determine which var. to be kept fixed during optimisation.
        """
        # possibly makes these properties of a class
        var_names = ['g', 'beta', 'logpsi', 'logphi', 'loggamma', 'logtau', 'logalpha']

        # controls the default fixing of parameters
        default = {'g': False, 'beta': True, 'logpsi': True,
                   'logphi': True, 'loggamma': False, 'logtau': True, 'logalpha': True}
        
        # check **kwargs to see if any variables have been kept fixed
        is_fixed_vars = [kwargs.pop("".join((vn, "_is_fixed")),
                                    default[vn])
                         for vn in var_names]

        return is_fixed_vars

    def _fit_init(self, is_fixed_vars, **kwargs):
        """
        Handles initialisation of optimisation
        """
        init_strategies = {
            'g': lambda : np.zeros(self.dim.N*self.dim.R),
            'logpsi': lambda : np.concatenate([gp.kernel.theta
                                               for gp in self.latentforces]),
            'beta': lambda : np.row_stack((np.zeros(self.dim.D),
                                           np.eye(self.dim.R, self.dim.D))).ravel(),
            'logphi': lambda : np.concatenate([gp.kernel.theta
                                               for gp in self.latentstates]),
            'loggamma': lambda : np.log(1e-4*np.ones(self.dim.K)),
            'logtau': lambda : np.log(1e4*np.ones(self.dim.K)),
            'logalpha': lambda : np.zeros(self.dim.K),
            }

        var_names = ['g', 'beta', 'logpsi', 'logphi', 'loggamma', 'logtau', 'logalpha']
        full_init = [kwargs.pop("".join((vn, '0')),
                                init_strategies[vn]())
                     for vn in var_names]
        full_init_shape = [item.size for item in full_init]

        free_vars, fixed_vars = ([], [])
        for item, boolean in zip(full_init, is_fixed_vars):
            if boolean:
                fixed_vars.append(item)
            else:
                free_vars.append(item)
        free_vars_shape = [item.size for item in free_vars]

        return np.concatenate(free_vars), free_vars_shape, fixed_vars

    def _fit_find_priors(self, **kwargs):
        priors = {}
        for vn in ['logpsi', 'beta', 'logtau', 'loggamma', 'logalpha']:
            key = '_'.join((vn, 'prior'))
            try:
                priors[vn] = kwargs[key].logpdf
            except KeyError:
                pass
        return priors

    def _var_mixer(self, free_vars, free_vars_shape, fixed_vars, is_fixed_vars):
        """ Utility function to mix the free vars and the fixed vars
        """
        free_vars = util._unpack_vector(free_vars, free_vars_shape)
        if is_fixed_vars is None:
            return free_vars
        else:
            full_vars = []
            ifree, ifixed = (0, 0)
            for b in is_fixed_vars:
                if b:
                    full_vars.append(fixed_vars[ifixed])
                    ifixed += 1
                else:
                    full_vars.append(free_vars[ifree])
                    ifree += 1
            return full_vars

class PredictMixin:
    """
    Extends the capabilties to behave like the mutliout GP
    """
    def predict_lf(self, tt, return_std=False):
        Cff_ = block_diag(*(gp.kernel_(tt[:, None], self.X_train_)
                            for gp in self.latentforces))
        lf_mean = Cff_.dot(self.g_alpha_)
        lf_mean = lf_mean.reshape(self.dim.R, tt.size).T

        if return_std:
            A = Cff_.dot(cho_solve((self.g_L_, True), np.eye(self.dim.N*self.dim.R)))
            K = A.dot(self.g_laplace_cov.dot(A.T))
            Cp = []
            for gp in self.latentforces:
                C22 = gp.kernel_(tt[:, None])
                C21 = gp.kernel_(tt[:, None], self.X_train_)
                Cp.append(C22 - C21.dot(cho_solve((gp.L_, True), C21.T)))
            K += block_diag(*Cp)
            return lf_mean, np.sqrt(np.diag(K))

        return lf_mean

class MLFMSuccApprox(PredictMixin, MLFMSAFitMixin, BaseMLFMSuccApprox):

    def __init__(self, *args, **kwargs):
        super(MLFMSuccApprox, self).__init__(*args, **kwargs)

    def fit(self, times, Y, **kwargs):
        """ Carries out MAP optimisation of the model parameters.
        """
        def objfunc(arg, free_vars_shape, fixed_vars):
            g, beta, logpsi, logphi, loggamma, logtau, logalpha = \
               self._var_mixer(arg, free_vars_shape, fixed_vars, is_fixed_vars)

            # reshape beta
            beta = beta.reshape(self.dim.R+1, self.dim.D)

            try:
                ll, ll_g_grad, ll_lgam_grad, ll_lalf_grad = \
                    self.log_likelihood(Y, g, beta,
                                        logphi, loggamma, logtau, logalpha,
                                        eval_gradient=True)
                lp, lp_g_grad = \
                    self.prior_logpdf(g, logpsi,
                                      eval_gradient=True)

                grad = [ll_g_grad + lp_g_grad,
                        None,
                        None,
                        None,
                        ll_lgam_grad,
                        None,
                        ll_lalf_grad]

                # add contributions from various priors
                for vn, indx in zip(['loggamma'],
                                    [(4, loggamma),]):
                    try:
                        prior_logpdf = priors[vn]
                        ind, x = indx  # unpack the value of var. name
                        vn_lp, vn_lp_grad = prior_logpdf(x, True)
                        lp += vn_lp
                        grad[ind] += vn_lp_grad
                    except KeyError:
                        pass

                grad = np.concatenate([item
                                       for item, b in zip(grad, is_fixed_vars)
                                       if not b])
                return -(ll+lp), -grad
                
            except:
                return np.inf, np.zeros(arg.shape)

        
        # make sure the model is ready for fitting by calling _setup
        self._setup(times, **kwargs)

        # parse  kwargs to see if any args to be kept fixed
        is_fixed_vars = self._fit_kwarg_parser(**kwargs)

        # Check for any priors
        priors = self._fit_find_priors(**kwargs)

        init, free_vars_shape, fixed_vars = \
              self._fit_init(is_fixed_vars, **kwargs)

        res = minimize(objfunc, init, jac=True,
                       args=(free_vars_shape, fixed_vars))

        # unpack the results
        g_, beta_, logpsi_, logphi_, loggamma_, logtau_, logalpha_ = \
            self._var_mixer(res.x, free_vars_shape, fixed_vars, is_fixed_vars)
        logpsi_ = util._unpack_vector(logpsi_,
                                      util._get_gp_theta_shape(self.latentforces))

        for r, gp in enumerate(self.latentforces):
            gp.kernel_ = gp.kernel.clone_with_theta(logpsi_[r])
            Cr = gp.kernel_(self.ttc[:, None])
            Cr[np.diag_indices_from(Cr)] += 1e-5
            gp.L_ = cholesky(Cr, True)

        
        # store some variables for later use
        self.X_train_ = self.ttc[:, None]
        if not is_fixed_vars[0]:
            # Optimised g - so get Laplace approx to from optim
            self.g_laplace_cov = res.hess_inv[:self.dim.N*self.dim.R,
                                              :self.dim.N*self.dim.R]
            Cff = block_diag(*(gp.kernel_(self.X_train_)
                               for gp in self.latentforces))
            Cff[np.diag_indices_from(Cff)] += 1e-4
            self.g_L_ = cholesky(Cff, lower=True)
            self.g_alpha_ = cho_solve((self.g_L_, True), g_)

        return MLFMSuccApproxFitResults(g_.reshape(self.dim.R, self.dim.N),
                                        beta_, logpsi_, logphi_, loggamma_, logtau_, logalpha_,
                                        res)

    def marginal_covar_matrix(self, g, beta, gamma, logphi, alpha, eval_gradient=False):
        """
        Covariance matrix K^n C0 K^nT
        """
        # reshape g
        g = g.reshape(self.dim.R, self.dim.N).T        # reshape 
        g = np.column_stack((np.ones(self.dim.N), g))  # and append a col. of 1s

        # get the struct matrices
        struct_mats = np.array([sum(brd*Ld for brd, Ld in zip(br, self.basis_mats))
                                for br in beta])
        # Add an axis for n=0,...,N-1
        At = struct_mats[None, :, :, :] * g[..., None, None]
        At = At.sum(1)

        # Get the additive noise error term
        #Gamma = np.diag(np.concatenate([gk*np.ones(self.dim.N) for gk in gamma]))
        Gamma = np.kron(np.eye(self.dim.N), np.diag(gamma))

        # Get the initial state covariance matrix
        C = np.kron(np.ones((self.dim.N, self.dim.N)),
                    np.diag(alpha))
        
        ####
        # Construct the transformation matrix K
        W = self._weight_matrix
        K = sum(sparse.kron(sparse.kron(W[:, i][:, None],
                                        sparse.eye(1, self.dim.N, i)),
                            At[i, ...])
                for i in range(self.dim.N))
        I = sparse.eye(self.dim.K)

        eifix = sparse.eye(1, self.dim.N, self.ifix)
        K += sum(sparse.kron(sparse.kron(sparse.eye(1, self.dim.N, i).transpose(),
                                         eifix),
                             I)
                 for i in range(self.dim.N))
        #_K = np.array(K.todense())
        if eval_gradient:
            dCdg = [np.zeros((self.dim.N*self.dim.K, self.dim.N*self.dim.K))]*self.dim.N*self.dim.R            

            #_dCdg = np.zeros((self.dim.N*self.dim.R,
            #                  self.dim.N*self.dim.K,
            #                  self.dim.N*self.dim.K))

            # Gradient of the operator K wrt g_rn
            dK = [sparse.kron(sparse.kron(W[:, n][:, None],
                                          sparse.eye(1, self.dim.N, n)),
                              struct_mats[r+1])
                  for r in range(self.dim.R)
                  for n in range(self.dim.N)]
            #_dK = np.array([np.kron(W[:, n][:, None], struct_mats[r+1])
            #                for r in range(self.dim.R)
            #                for n in range(self.dim.N)])

            dCdgamma = np.zeros((self.dim.N*self.dim.K,
                                 self.dim.N*self.dim.K,
                                 self.dim.K))
            dCdalpha = [np.kron(np.ones((self.dim.N, self.dim.N)),
                                np.diag(np.eye(N=1, M=self.dim.K, k=k).ravel()))
                        for k in range(self.dim.K)]

            for m in range(self.order):

                dCdg = np.array([K.dot(K.dot(dC_grn).T) for dC_grn in dCdg])          
                for i, dK_grn in enumerate(dK):
                    kcdk = K.dot(dK_grn.dot(C).T)
                    dCdg[i] += kcdk + kcdk.T

                # update gradient of C wrt to g_{rn}
                #CKt = K.dot(C).T
                #print(K.shape, _dCdg.shape)
                #expr = _K.dot(_dCdg)
                #print(expr.shape)
                #assert(False)
                #expr = dK.dot(CKt.reshape(self.dim.N, self.dim.K, self.dim.N*self.dim.K)).sum(2)
                #expr += expr.transpose(0, 2, 1)
                #expr2 = K.dot(K.dot(dCdg).T)
                #dCdg = expr + expr2
                #dCdg = [K.dot(dK_grn.dot(C).T) + \
                #        dK_grn.dot(K.dot(C).T) + \
                #        K.dot(K.dot(dC_grn).T)
                #        for dK_grn, dC_grn in zip(dK, dCdg)]

                # update gradient of C wrt to Gamma
                dCdgamma = np.dstack((K.dot(K.dot(dCdgamma[..., i]).T)
                                      for i in range(dCdgamma.shape[-1])))

                diag_ek = np.zeros((self.dim.K, self.dim.K))
                for k in range(self.dim.K):
                    diag_ek[k, k] = 1.
                    dCdgamma[..., k] += np.kron(np.eye(self.dim.N), diag_ek)
                    diag_ek[k, k] = 0.

                # update gradient of C wrt to alpha
                dCdalpha = [K.dot(K.dot(dCdak).T) for dCdak in dCdalpha]
                
                C = K.dot(K.dot(C).T) + Gamma

            #print("...done.")
            dCdg = np.dstack(dCdg)
            return (C,
                    dCdg,
                    dCdgamma,
                    np.dstack(dCdalpha))

        else:
            for m in range(self.order):
                C = K.dot(K.dot(C).T) + Gamma            
            return C

    def marginal_covar_matrix2(self, g, beta, gamma, logphi, alpha, eval_gradient=False):
        """
        Rewrite to prevent some of the looping
        """
        g = g.reshape(self.dim.R, self.dim.N).T        # reshape 
        g = np.column_stack((np.ones(self.dim.N), g))  # and append a col. of 1s

        # get the struct matrices
        struct_mats = np.array([sum(brd*Ld for brd, Ld in zip(br, self.basis_mats))
                                for br in beta])
        # Add an axis for n=0,...,N-1
        At = struct_mats[None, :, :, :] * g[..., None, None]
        At = At.sum(1)

        # Get the additive noise error term
        Gamma = np.kron(np.eye(self.dim.N), np.diag(gamma))

        # Get the initial state covariance matrix
        C = np.kron(np.ones((self.dim.N, self.dim.N)),
                    np.diag(alpha))
        
        ####
        # Construct the transformation matrix K
        W = self._weight_matrix
        K = sum(sparse.kron(sparse.kron(W[:, i][:, None],
                                        sparse.eye(1, self.dim.N, i)),
                            At[i, ...])
                for i in range(self.dim.N))
        I = sparse.eye(self.dim.K)

        eifix = sparse.eye(1, self.dim.N, self.ifix)
        K += sum(sparse.kron(sparse.kron(sparse.eye(1, self.dim.N, i).transpose(),
                                         eifix),
                             I)
                 for i in range(self.dim.N))
        
        if eval_gradient:
            # dK[r, n, ...] = wn (x) Ar
            dK = [sparse.kron(sparse.kron(W[:, n][:, None],
                                          sparse.eye(1, self.dim.N, n)),
                              struct_mats[r+1])
                  for r in range(self.dim.R)
                  for n in range(self.dim.N)]

            dCdg = np.zeros((self.dim.N*self.dim.R,
                             self.dim.N*self.dim.K,
                             self.dim.N*self.dim.K))

            dCdalpha = [np.kron(np.ones((self.dim.N, self.dim.N)),
                                np.diag(np.eye(N=1, M=self.dim.K, k=k).ravel()))
                        for k in range(self.dim.K)]

            K = K.todense()
            Kt = np.array(K.transpose())
            
            for m in range(self.order):
                CKt = C.dot(Kt)

                # Update gradient of C wrt to g_{rn}
                for i, (dK_rn, dC_grn) in enumerate(zip(dK, dCdg)):
                    a = dK_rn.dot(CKt)
                    dCdg[i] = a + a.T + K.dot(dC_grn.dot(Kt))

                # Update gradient of C wrt to alpha
                dCdalpha = [K.dot(dCdak.dot(Kt))
                            for dCdak in dCdalpha]

                # Finally update C
                C = K.dot(CKt) + Gamma

            return C, dCdg, np.stack(dCdalpha)


    def data_cov(self, tau, logphi):
        Kdat = np.diag((np.ones((self.Ndata, self.dim.K))*(1/tau)).ravel())
        return Kdat
        logphi = util._unpack_vector(logphi,
                                     util._get_gp_theta_shape(self.latentstates))
        if self.is_tt_aug:
            tt = self._tt_data
        else:
            tt = self.ttc
        datacov = [gp.kernel.clone_with_theta(theta)(tt[:, None])
                   for gp, theta in zip(self.latentstates, logphi)]
        for item in datacov:
            item /= tau[0]
        datacov = block_diag(*datacov)
        #for k, C in enumerate(datainvcov):
        #    C[np.diag_indices_from(C)] += self.latentstates[k].alpha
        #datainvcov = [np.linalg.cholesky(C) for C in datainvcov]
        #datainvcov = block_diag(*[cho_solve((L, True), np.eye(L.shape[0]))
        #for L in datainvcov])
        # this is the inverse of Cov{vec(X)}, we need
        # Cov{rav(X)} = T Cov{vec(X)} Tt
        #   ---> InvCov{rav(X)} = Tt InvCov{vec(X)} T
        T = util.T_xrav_tovecx(self.Ndata, self.dim.K)
        #datainvcov = Tt.dot(Tt.dot(datainvcov).T)
        #Kdat += T.dot(T.dot(datacov).T)
        Kdat = datacov

        return Kdat
        
    def log_likelihood(self, y, g, beta,
                       logphi, loggamma, logtau, logalpha,
                       eval_gradient=False, **kwargs):
        if eval_gradient:
            K, K_g_grad, K_gamma_grad, K_alpha_grad = \
               self.marginal_covar_matrix(g, beta, np.exp(loggamma), logphi, np.exp(logalpha), True)
        else:
            K = self.marginal_covar_matrix(g, beta, np.exp(loggamma), logphi, np.exp(logalpha))

        if self.is_tt_aug:
            data_inds = np.asarray(self.data_inds)
            data_inds = (data_inds*self.dim.K)[:, None] + np.arange(self.dim.K)[None, :]
            data_inds_ix_ = np.ix_(data_inds.ravel(), data_inds.ravel())

            K = K[data_inds_ix_]
            if eval_gradient:
                K_g_grad = np.dstack([K_g_grad[..., i][data_inds_ix_]
                                      for i in range(K_g_grad.shape[-1])])
                K_gamma_grad = np.dstack([K_gamma_grad[..., i][data_inds_ix_]
                                          for i in range(K_gamma_grad.shape[-1])])
                K_alpha_grad = np.dstack([K_alpha_grad[..., i][data_inds_ix_]
                                          for i in range(K_alpha_grad.shape[-1])])

        # data covariance
        tau = np.exp(logtau)
        Kdat = self.data_cov(tau, logphi)
        K += Kdat

        try:
            L = np.linalg.cholesky(K)
        except np.linalg.LinAlgError:
            return (-np.inf, np.zeros_like(g.size)) \
                   if eval_gradient else -np.inf

        if len(y.shape) == 1:
            y_train = y[:, None]
        else:
            y_train = y
        
        alpha = cho_solve((L, True), y_train)

        log_lik_dims = -.5 * np.einsum("ik, ik->k",y_train, alpha)
        log_lik_dims -= np.log(np.diag(L)).sum()
        log_lik_dims -= K.shape[0] / 2 * np.log(2 * np.pi)
        log_lik = log_lik_dims.sum(-1)  # sum over output dimensions

        if eval_gradient:
            tmp = np.einsum("ik,jk->ijk", alpha, alpha)
            tmp -= cho_solve((L, True), np.eye(K.shape[0]))[:, :, None]

            log_likelihood_gradient_dims = \
                0.5 * np.einsum("ijl,ijk->kl", tmp, K_g_grad)
            log_likelihood_gradient = log_likelihood_gradient_dims.sum(-1)


            log_likelihood_gamma_gradient_dims = \
                0.5 * np.einsum("ijl,ijk->kl", tmp, K_gamma_grad)
            log_likelihood_loggamma_gradient = log_likelihood_gamma_gradient_dims.sum(-1)
            # Gamma parameters is on the log scale
            log_likelihood_loggamma_gradient = \
                log_likelihood_loggamma_gradient * np.exp(loggamma)

            log_likelihood_alpha_gradient_dims = \
                0.5 * np.einsum('ijl,ijk->kl', tmp, K_alpha_grad)
            log_likelihood_logalpha_gradient = log_likelihood_alpha_gradient_dims.sum(-1)
            # alpha parameters are on the log scale
            log_likelihood_logalpha_gradient = \
                log_likelihood_logalpha_gradient * np.exp(logalpha)

            return (log_lik,
                    log_likelihood_gradient,
                    log_likelihood_loggamma_gradient,
                    log_likelihood_logalpha_gradient)
        else:
            return log_lik

    def prior_logpdf(self, g, logpsi, eval_gradient):
        """Logpdf of prior
        """
        veclogpsi = logpsi.copy()
        logpsi_shape = util._get_gp_theta_shape(self.latentforces)
        logpsi = util._unpack_vector(logpsi, logpsi_shape)

        # reshapge g
        g = g.reshape(self.dim.R, self.dim.N)
        ll = 0.

        if eval_gradient:
            ll_g_grad, ll_logpsi_grad = ([], [])

        for r in range(self.dim.R):
            gp = self.latentforces[r]
            kern = gp.kernel.clone_with_theta(logpsi[r])

            if eval_gradient:
                K, K_gradient = kern(self.ttc[:, None],
                                     eval_gradient=True)
            else:
                K = kern(self.ttc[:, None])

            K[np.diag_indices_from(K)] += gp.alpha
            L = np.linalg.cholesky(K)

            alpha = cho_solve((L, True), g[r, :])

            tmp = -.5 * g[r, :].dot(alpha)
            tmp -= np.log(np.diag(L)).sum()

            ll += -.5 * g[r, :].dot(alpha) - \
                  np.log(np.diag(L)).sum() - \
                  L.shape[0] / 2 * np.log(2 * np.pi)

            if eval_gradient:
                ll_g_grad.append(-alpha)

        if eval_gradient:
            return ll, \
                   np.concatenate(ll_g_grad)

        return ll


class VarMLFMSuccApprox(BaseMLFMSuccApprox):

    def varfit(self, times, Y, gtol=1e-3, max_iter=100, **kwargs):

        try:
            mapres = kwargs.pop('mapres')
            self._setup(times, **kwargs)
        except KeyError:
            raise NotImplementedError("varfit currently requires the results of a MAP fit")

        # initalise the distribution of g at the MAP
        Eg = mapres.g.ravel()
        Covg = np.diag([0.1]*Eg.size)

        Eb = mapres.beta.ravel()

        Deltag = np.inf

        # initalise the distribution of X
        # X is of shape[N*K, order+1, n_samples]
        Ex, Covx = self._init_x_distrib(times, Y)
        # Burn in the distribution of Ex, Covx around g map

        for i in range(5):
            Ex, Covx = self.x_update_moments(Y, Ex, Covx, Eg, Covg,
                                             mapres.beta, np.exp(mapres.loggamma),
                                             mapres.logphi, mapres.logtau)
        
        nt = 0  # Count number of iterations
        while nt < 10:
            Eg, Covg = self.g_update_moments(Ex, Covx,
                                             mapres.beta, np.exp(mapres.loggamma),
                                             mapres.logpsi)            
            Ex, Covx = self.x_update_moments(Y,
                                             Ex, Covx,
                                             Eg, Covg,
                                             mapres.beta, np.exp(mapres.loggamma),
                                             mapres.logphi, mapres.logtau)
            nt += 1

        return Eg, Covg

    def _init_x_distrib(self, times, Y):
        from scipy.interpolate import interp1d
        if self.is_tt_aug:
            # cheap interpolation of the data to the latent variable times
            Ex = np.zeros((self.dim.N*self.dim.K, self.order+1, Y.shape[1]))
            for m, Ym in enumerate(Y.T):
                Ym = Ym.reshape(self.Ndata, self.dim.K)
                u = [interp1d(times, ymk,
                              kind='cubic',
                              fill_value='extrapolate')
                     for ymk in Ym.T]

                Yinterp = np.column_stack([uk(self.ttc)
                                           for uk in u]).ravel()
                Ex[:, :, m] = np.column_stack((Yinterp
                                               for p in range(self.order+1)))
            NK = self.dim.N*self.dim.K
            
            Covx = np.dstack([np.diag(0.001*np.ones(NK))
                              for p in range(min(self.order, 3))])
            return Ex, Covx

    def g_update_moments(self, EX, CovX, beta, gamma, logpsi):

        # dimensionality variables
        n_outputs = EX.shape[-1]
        NK = self.dim.N*self.dim.K

        # struct mats
        struct_mats = np.array([sum(brd*Ld for brd, Ld in zip(br, self.basis_mats))
                                for br in beta])

        # vec(K) = v0 + V.dot(g)
        V, v0 = self.sparse_vecK_aff_rep(beta)
        V = V.todense()
        v0 = v0.todense()

        GammaInv = np.diag(np.concatenate([(1/gk)*np.ones(self.dim.N)
                                           for gk in gamma]))

        # array of outer products
        EXXT = np.einsum('ikl,jkl->ijkl', EX[...,:-1,:], EX[...,:-1,:])
        # ... adjust with covariance matrices
        EXXT[..., 0, :] += CovX[..., 0][..., None]
        if EXXT.shape[1] > 2:
            EXXT[..., 1:-1, :] += CovX[..., 1][..., None, None]
            EXXT[..., -1, :] += CovX[..., 2][..., None]
        else:
            EXXT[..., -1, :] += CovX[..., 1]
        S = np.kron(EXXT.sum(2), GammaInv[..., None])

        ## E outer(Xm, X_{m+1})
        # - we use the fact cov{X_m, X_{m+1}} = 0 under chosen mean field factorisation
        EXXpT = np.einsum('ikl,jkl->ijkl', EX[..., 1:, :], EX[..., :-1, :]).sum(2)

        # b = vec(sum([Gamma.dot(outer(xm, xm-1)) ]))
        b = np.einsum('ij,jkl->ikl', GammaInv, EXXpT)
        b = b.transpose(1, 0, 2).reshape(NK*NK, n_outputs)

        premean = np.einsum('ijl,jk->ikl', S, v0)
        premean = b[:, None, :] - premean
        premean = np.einsum('ij,jkl->ikl', V.T, premean)
        
        invcov = np.einsum('ijl,jk->ikl', S, V)
        invcov = np.einsum('ij,jkl->ikl', V.T, invcov)

        invcov = invcov.sum(-1)
        # get the contribution from the prior
        logpsi = util._unpack_vector(logpsi,
                                     util._get_gp_theta_shape(self.latentforces))
        for r, theta in enumerate(logpsi):
            kern = self.latentforces[r].kernel.clone_with_theta(theta)
            K = kern(self.ttc[:, None])
            K[np.diag_indices_from(K)] += self.latentforces[r].alpha
            L = np.linalg.cholesky(K)
            invcov[r*self.dim.N:(r+1)*self.dim.N,
                   r*self.dim.N:(r+1)*self.dim.N] += \
                   cho_solve((L, True), np.eye(L.shape[0]))

        L = np.linalg.cholesky(invcov)
        cov = cho_solve((L, True), np.eye(L.shape[0]))
        mean = cov.dot(premean.sum(-1))
        #mean = np.einsum('ij,jkl->ikl', cov, premean.sum(-1))[:, 0].sum(-1)

        return mean.ravel(), cov

    def x_update_moments(self, Y, EX, CovX, Eg, Covg,
                         beta, gamma, logphi, logtau):

        # Get the additive noise error term
        GammaInv = np.diag(
            np.column_stack([1/gk*np.ones(self.dim.N)
                             for gk in gamma]).ravel())
        
        # x0 cov variance matrix
        logphi = util._unpack_vector(logphi,
                                     util._get_gp_theta_shape(self.latentstates))
        C0inv = [gp.kernel.clone_with_theta(theta)(self.ttc[:, None])
                 for gp, theta in zip(self.latentstates, logphi)]
        for k, C in enumerate(C0inv):
            C[np.diag_indices_from(C)] += self.latentstates[k].alpha
        C0inv = [np.linalg.cholesky(C) for C in C0inv]
        C0inv = block_diag(*[cho_solve((L, True), np.eye(L.shape[0]))
                             for L in C0inv])
        # this is the inverse of Cov{vec(X)}, we need
        # Cov{rav(X)} = T Cov{vec(X)} Tt
        #   ---> InvCov{rav(X)} = Tt InvCov{vec(X)} T
        Tt = util.T_xrav_tovecx(self.dim.N, self.dim.K).transpose()
        C0inv = Tt.dot(Tt.dot(C0inv).T)

        ## Takes the expectation wrt to distribution of G
        EK = self._K(Eg, beta)
        EKt = EK.transpose()
        EKtLK = self._EKtLK(Eg, Covg, beta, GammaInv)

        ######
        # q(X0)
        ic0 = C0inv + EKtLK
        pm0 = EKt.dot(GammaInv.dot(EX[:, 1, :]))
        L0 = np.linalg.cholesky(ic0)
        cov0 = cho_solve((L0, True), np.eye(L0.shape[0]))
        EX[:, 0, :] = cov0.dot(pm0)
        CovX[..., 0] = cov0

        # Common cov
        cov = GammaInv + EKtLK
        L = np.linalg.cholesky(cov)
        cov = cho_solve((L, True), np.eye(L.shape[0]))
        CovX[..., 1] = cov        
        for m in range(self.order-1):
            pm = EKt.dot(GammaInv.dot(EX[:, m+2, :])) + \
                 GammaInv.dot(EK.dot(EX[:, m, :]))
            EX[:, m+1, :] = cov.dot(pm)

        datainvcov = np.diag(
            np.column_stack([tau*np.ones(self.Ndata)
                             for tau in np.exp(logtau)]).ravel())

        datainvcov = [gp.kernel.clone_with_theta(theta)(self._tt_data[:, None])
                 for gp, theta in zip(self.latentstates, logphi)]
        for k, C in enumerate(datainvcov):
            C[np.diag_indices_from(C)] += self.latentstates[k].alpha
        datainvcov = [np.linalg.cholesky(C) for C in datainvcov]
        datainvcov = block_diag(*[cho_solve((L, True), np.eye(L.shape[0]))
                             for L in datainvcov])
        # this is the inverse of Cov{vec(X)}, we need
        # Cov{rav(X)} = T Cov{vec(X)} Tt
        #   ---> InvCov{rav(X)} = Tt InvCov{vec(X)} T
        Tt = util.T_xrav_tovecx(self.Ndata, self.dim.K).transpose()
        datainvcov = Tt.dot(Tt.dot(datainvcov).T)

        invcov = GammaInv
        if self.is_tt_aug:
            data_inds = np.asarray(self.data_inds)
            data_inds = (data_inds*self.dim.K)[:, None] + \
                        np.arange(self.dim.K)[None, :]
            data_inds_ix_ = np.ix_(data_inds.ravel(), data_inds.ravel())
            invcov[data_inds_ix_] += datainvcov
            pm = GammaInv.dot(EK.dot(EX[:, -2, :]))
            pm[data_inds.ravel(), :] += datainvcov.dot(Y)
        else:
            invcov += datainvcov
            pm = datainvcov.dot(Y) + GammaInv.dot(EK.dot(EX[:, -2, :]))

        cov = np.linalg.inv(invcov) #np.diag(1 / np.diag(invcov))
        
        EX[:, -1, :] = cov.dot(pm)
        CovX[..., -1] = cov

        return EX, CovX

    def _EKtLK(self, Eg, Covg, beta, L):
        """
        Computes the expected value of the matrix valued quadratic form.
        """
        EggT = np.outer(Eg, Eg) + Covg

        # vector v0 and matrix V with vec(K) = v0 + V.dot(g)
        V, v0 = self.sparse_vecK_aff_rep(beta)
        EVg = V.dot(Eg[:, None])  # E[ V.dot(g) ] = V.dot(Eg)
        vEVg = v0.dot(EVg.T)      # E[ outer(v, Vg) ] = outer(v, V.dot(Eg))

        # E[outer(vec(K), vec(K)]
        E_vecK_outer = v0.dot(v0.transpose()) + \
                       vEVg + vEVg.T + \
                       V.dot(V.dot(EggT).T)

        # reshape so that [n, m, ...] = E[outer(K[:, n], K[:, m])]
        NK = self.dim.N * self.dim.K
        # numpy matrix to array
        E_vecK_outer = np.asarray(E_vecK_outer)
        E_vecK_outer = E_vecK_outer.reshape(NK, NK, NK, NK).transpose(0, 2, 1, 3)
        # reshape so that [n, m, ...] = vec(E[outer(K[:, m], K[:, m])])
        E_vecK_outer = E_vecK_outer.transpose(0, 1, 3, 2).reshape(NK, NK, NK**2)

        # final results is [n, m] = Tr(E_vecK_outer[n, m, ...].dot(L))
        #   i) vectorise L
        #  ii) multiply and sum for dot product
        return (E_vecK_outer * L.T.ravel()[None, None, :]).sum(-1)

"""
Methods for trapezoidal integration
"""
def get_weights(tt):
    tt = np.asarray(tt)
    W = np.zeros((tt.size, tt.size))
    h = np.diff(tt)
    for i in range(tt.size):
        W[i, :i] += .5*h[:i]
        W[i, 1:i+1] += .5*h[:i]
    return W

def weight_matrix(tt, ifix):

    ttb = tt[:ifix+1]
    ttf = tt[ifix:]

    Wb = get_weights([t for t in reversed(ttb)])
    Wb = np.flip(Wb, (0, 1))
    Wf = get_weights(ttf)

    W = np.zeros((tt.size, tt.size))
    W[:ifix+1, :ifix+1] += Wb
    W[ifix:, ifix:] += Wf

    return W



import numpy as np
import sklearn.gaussian_process.kernels as sklearn_kernels
from scipy.spatial.distance import pdist, cdist, squareform

def _check_length_scale(X, length_scale):
    length_scale = np.squeeze(length_scale).astype(float)
    if np.ndim(length_scale) > 1:
        raise ValueError("length_scale cannot be of dimension greater than 1")
    if np.ndim(length_scale) == 1 and X.shape[1] != length_scale.shape[0]:
        raise ValueError("Anisotropic kernel must have the same number of "
                         "dimensions as data (%d!=%d)"
                         % (length_scale.shape[0], X.shape[1]))
    return length_scale

class GradientKernel(sklearn_kernels.Kernel):
    """
    Base class for the gradient kernel.
    """
    def __mul__(self, b):
        if isinstance(b, GradientKernel):
            return GradientKernelProduct(self, b)
        else:
            raise ValueError("Multiplication must be between two GradientKernels")

    def __add__(self, b):
        raise NotImplementedError("Addition of GradientKernels not yet supported. Feel free to contribute!")

class GradientKernelProduct(sklearn_kernels.Product):

    def __call__(self, X, Y=None, eval_gradient=False, comp='x'):
        if comp == 'x':
            return super(GradientKernelProduct, self).__call__(X, Y, eval_gradient=eval_gradient)

        elif comp == 'xdx':
            if eval_gradient:
                K1, K1_gradient = self.k1(X, Y, eval_gradient=True)
                K1dx, K1dx_gradient = self.k1(X, Y, comp='xdx', eval_gradient=True)
                K2, K2_gradient = self.k2(X, Y, eval_gradient=True)
                K2dx, K2dx_gradient = self.k2(X, Y, comp='xdx', eval_gradient=True)

                # gradient wrt first par.
                grad1 = K1dx_gradient * K2[..., np.newaxis, np.newaxis] + \
                        K1_gradient[...,np.newaxis, :] * K2dx[..., np.newaxis]

                # gradient wrt second par.
                grad2 = K1dx[..., np.newaxis] * K2_gradient[..., np.newaxis, :] + \
                        K1[..., np.newaxis, np.newaxis] * K2dx_gradient

                Kdx = K1dx * K2[..., np.newaxis] + K1[..., np.newaxis] * K2dx
                Kdx_gradient = np.stack((grad1, grad2), axis=3)

                return Kdx, Kdx_gradient[...,0]
                
            else:
                K1 = self.k1(X, Y)
                K1dx = self.k1(X, Y, comp='xdx')
                K2 = self.k2(X, Y)
                K2dx = self.k2(X, Y, comp='xdx')
                return K1dx * K2[..., np.newaxis] + K1[..., np.newaxis] * K2dx

        elif comp == 'dxdx':
            if eval_gradient:
                K1, K1_gradient = self.k1(X, Y, eval_gradient=True)
                K1dx, K1dx_gradient = self.k1(X, Y, comp='xdx', eval_gradient=True)
                K1dxdx, K1dxdx_gradient = self.k1(X, Y, comp='dxdx', eval_gradient=True)
                K2, K2_gradient = self.k2(X, Y, eval_gradient=True)
                K2dx, K2dx_gradient = self.k2(X, Y, comp='xdx', eval_gradient=True)
                K2dxdx, K2dxdx_gradient = self.k2(X, Y, comp='dxdx', eval_gradient=True)

                grad1 = K1dxdx_gradient * K2[..., np.newaxis, np.newaxis, np.newaxis] + \
                        K1dx_gradient[:, :, np.newaxis, :, :] * \
                        K2dx[:, :, :, np.newaxis, np.newaxis] + \
                        K1dx_gradient[:, :, :, np.newaxis, :] * \
                        K2dx[:, :, np.newaxis, :, np.newaxis] + \
                        K1_gradient[..., np.newaxis, np.newaxis, :] * \
                        K2dxdx[..., np.newaxis]

                grad2 = K1dxdx[..., np.newaxis] * \
                        K2_gradient[..., np.newaxis, np.newaxis] + \
                        K1dx[:, :, np.newaxis, :, np.newaxis] * \
                        K2dx_gradient[:, :, :, np.newaxis, :] + \
                        K1dx[:, :, :, np.newaxis, np.newaxis] * \
                        K2dx_gradient[:, :, np.newaxis, :, :] + \
                        K1[..., np.newaxis, np.newaxis, np.newaxis] * \
                        K2dxdx_gradient

                Kdxdx_gradient = np.stack((grad1, grad2), axis=4)

                Kdxdx = K1dxdx * K2[..., np.newaxis, np.newaxis] + \
                        K1dx[..., np.newaxis, :] * K2dx[..., np.newaxis] + \
                        K1dx[..., np.newaxis] * K2dx[..., np.newaxis, :] + \
                        K1[..., np.newaxis, np.newaxis] * K2dxdx

                return Kdxdx, Kdxdx_gradient[..., 0]



            else:
                K1 = self.k1(X, Y)
                K1dx = self.k1(X, Y, comp='xdx')
                K1dxdx = self.k1(X, Y, comp='dxdx')
                K2 = self.k2(X, Y)
                K2dx = self.k2(X, Y, comp='xdx')
                K2dxdx = self.k2(X, Y, comp='dxdx')

                return K1dxdx * K2[..., np.newaxis, np.newaxis] + \
                       K1dx[..., np.newaxis, :] * K2dx[..., np.newaxis] + \
                       K1dx[..., np.newaxis] * K2dx[..., np.newaxis, :] + \
                       K1[..., np.newaxis, np.newaxis] * K2dxdx

class ConstantKernel(GradientKernel, sklearn_kernels.ConstantKernel):
    """Constant kernel.

    Can be used as part of a product-kernel where it scales the magnitude of
    the other factor (kernel) or as part of a sum-kernel, where it modifies
    the mean of the Gaussian process.

    For the gradient kernel higher order derivatives are zero.

    k(x_1, x_2) = constant_value for all x_1, x_2

    Parameters
    ----------
    constant_value : float, default: 1.0
        The constant value which defines the covariance:
        k(x_1, x_2) = constant_value

    constant_value_bounds : pair of floats >= 0, default: (1e-5, 1e-5)
        The lower and upper bound on constant_value

    """
    def __call__(self, X, Y=None, eval_gradient=False, comp='x'):
        if comp == 'x':
            return super(ConstantKernel, self).__call__(X, Y, eval_gradient)
        else:
            if Y is None:
                Y = X
            elif eval_gradient:
                raise ValueError

            if comp == 'xdx':
                if eval_gradient:
                    res = np.zeros((X.shape[0], X.shape[0], X.shape[1]))
                    return res, res[..., np.newaxis]
                else:
                    return np.zeros((X.shape[0], Y.shape[0], X.shape[1]))

            elif comp == 'dxdx':
                if eval_gradient:
                    res = np.zeros((X.shape[0], X.shape[0],
                                    X.shape[1], X.shape[1],))
                    return res, res[..., np.newaxis]
                else:
                    return np.zeros((X.shape[0], Y.shape[0],
                                     X.shape[1], X.shape[1]))

class RBF(GradientKernel, sklearn_kernels.RBF):

    def __call__(self, X, Y=None, eval_gradient=False, comp='x'):
        if comp == 'x':
            return super(RBF, self).__call__(X, Y, eval_gradient)

        else:


            X = np.atleast_2d(X)
            length_scale = _check_length_scale(X, self.length_scale)

            if Y is None:
                Y = X

            Diffs = (X[:, np.newaxis, :] - Y[np.newaxis, :, :])

            if eval_gradient:
                K, K_gradient = super(RBF, self).__call__(X, eval_gradient=True)
            else:
                K = np.exp(-.5 * np.sum(Diffs ** 2 / (length_scale ** 2), axis=2) )

            if comp == 'xdx':

                Kdx = (Diffs / (length_scale ** 2)  ) * K[..., np.newaxis]

                if eval_gradient:
                    Diffs /= (length_scale ** 2)
                    K, K_gradient = super(RBF, self).__call__(X, None, eval_gradient=True)

                    Kdx_gradient = Diffs[..., np.newaxis] * K_gradient[..., np.newaxis, :] 

                    # add a diagonal term contribution
                    for d in range(Kdx_gradient.shape[-1]):
                        Kdx_gradient[..., d, d] -= 2*Diffs[..., d]*K

                    return Kdx, Kdx_gradient

                else:
                    return Kdx

            elif comp == 'dxdx':

                Diffs /= (length_scale ** 2)
                Kdxdx = -Diffs[..., np.newaxis, :] * Diffs[..., np.newaxis]

                # add a diagonal term contribution
                if Diffs.shape[-1] == 1:
                    Kdxdx[:, :, 0, 0] += 1 / (length_scale ** 2)
                    
                else:
                    for d in range(Diffs.shape[-1]):
                        Kdxdx[:, :, d, d] += 1 / (length_scale[d] ** 2)

                if eval_gradient:
                    # Kdkdx = f * K
                    P = Diffs.shape[-1]
                    if P == 1:
                        length_scale = [length_scale]
                    
                    f = Kdxdx.copy()
                    f_gradient = np.zeros(Kdxdx.shape + (P, ))

                    for p in range(P):
                        f_gradient[..., p, p, p] -= 2 / (length_scale[p]**2)
                        for i in range(P):
                            f_gradient[...,p, i, p] += 2*Diffs[...,p] * \
                                                       Diffs[..., i]
                            f_gradient[...,i, p, p] += 2*Diffs[...,p] * \
                                                       Diffs[..., i]
                    expr1 = f_gradient * K[..., np.newaxis, np.newaxis, np.newaxis]

                    expr2 = f[..., np.newaxis] * \
                            K_gradient[:, :, np.newaxis, np.newaxis, :]

                    return Kdxdx * K[..., np.newaxis, np.newaxis], \
                           expr1 + expr2

                else:
                    return Kdxdx * K[..., np.newaxis, np.newaxis]



from . import gradientkernels
from . import linlatentforcemodels
from . import liealgebras
from . import probabilitydistributions
from . import gaussian_process
from . import kernels
from . import pydygp_examples



import numpy as np
from . import util
from .mlfm import BaseMLFM, Dimensions
from scipy.linalg import block_diag, cho_solve
from scipy.stats import norm
from scipy.optimize import minimize
import scipy.sparse as sparse

def _unpack_vector(x, shape):
    res = []
    ntot = 0
    for n in shape:
        res.append(x[ntot:ntot+n])
        ntot += n
    return res

def Mpower_action_wgrad(M, dM, x, power, dMdb=None):
    """
    Returns the result of M^{power}x and its gradient. 
    """
    res = x
    grad = [np.zeros(x.size).reshape(x.shape)]*dM.shape[0]

    if dMdb is not None:
        grad_b = [np.zeros(x.size).reshape(x.shape)]*dMdb.shape[0]
    
    for i in range(power):
        grad = [dM_i.dot(res) + M.dot(grad_i)
                for dM_i, grad_i in zip(dM, grad)]
        if dMdb is not None:
            grad_b = [dM_r.dot(res) + M.dot(grad_i)
                      for dM_i, grad_i in zip(dMdb, grad_b)]
            
        res = M.dot(res)

    if dMdb is None:
        return res, np.array(grad)
    else:
        return res, grad, grad_b

def M2():

    # make K
    
    res = x
    grad = np.empty((N_repl, N, R, NK))
    for i in range(power):
        res, dres = KAction_grad()
        K_dot_grad = np.einsum('ij,mnrj->mnri', W, grad)
        grad += dres + K.dot(grad)
    return res, grad
        

def KAction_grad(x, g, beta, W,
                 N_repl, N, K, R):

    # X[m, k, ...] = kth dimension component of x
    X = x.T.reshape((N_repl, K, N))
    X = X.transpose((0, 2, 1))

    Arr = np.dstack([sum(brd*Ld
                         for brd, Ld in zip(br, self.basis_mats)).T
                     for br in beta[1:, :]])

    # XA[n, ..., r] = X[n, ...].dot(Arr[..., r])
    XA = np.einsum('ijk,klr->ijlk', X, Arr)

    # store the outer product of the nth col. W
    # with the nth row X[, n, ]

    # broadcasting:
    #    W        XA
    #  (N, N)   (M, N, K, R)
    #
    # WXA[m, n, :, k, r] = outer prod of the nth col. of W with
    # the mth replicate of X
    WXA = (W.T)[None, :, :, None, None] * XA[..., None, :, :]

    # now use the gradient to also construct K[g].dot(x) = K[0].dot(x) + sum( g * dKdg )

    # finally transpose and the reshape so that X has been vectorised
    return WXA.reshape((N_repl, N, R, N*K))

def Kpow_Action_grad(x, beta, power, W,
                     N_repl, N, K, R):
    res = x


# Utility functions for padding the time vector
def augment_times(t, h):
    """
    Pads the vector t so that the maximum spacing is h.

    Parameters
    ----------

    t : one dimensional array like

    h : float
    """
    inds = [0]
    res = [t[0]]
    for ta, tb in zip(t[:-1], t[1:]):
        N = np.ceil((tb - ta) / h + 1)
        N = int(N)
        _tt = np.linspace(ta, tb, N)
        res = np.concatenate((res, _tt[1:]))
        inds.append(inds[-1] + N - 1)
    return res, np.array(inds)    

def handle_time_inds(times, h=None):
    """
    Takes a list of time vectors and returns the unique, potentially augmented,
    vector 
    """
    # get size of each vector
    t_sizes = [len(t) for t in times]

    # concatenate to single time vector
    tt = np.concatenate(times)

    # get the distinct times, and the indices
    tt_uni, inv_ind = np.unique(tt, return_inverse=True)

    # split inv_ind up
    ind_ti = _unpack_vector(inv_ind, t_sizes)

    if h is None:
        return tt_uni, ind_ti

    elif isinstance(h, float) and h > 0:
        # augment the time vector so that diff is at most h
        ttc, inds_c = augment_times(tt_uni, h)
        data_inds = [inds_c[ind_i] for ind_i in ind_ti]
        return ttc, data_inds

    else:
        raise ValueError("h should be a float > 0")

class Softmax:
    def __init__(self, basis_funcs):
        self.basis_funcs = basis_funcs
        self.param_shape = [len(funcs) for funcs in basis_funcs]

    def __call__(self, v, X):
        if len(v) != sum(self.param_shape):
            raise ValueError("v should be an array like of"
                             " shape {}".format(sum(self.param_shape)))
        else:
            v = util._unpack_vector(v, self.param_shape)

        F = [np.column_stack((f(X) for f in bfuncs))
             for bfuncs in self.basis_funcs]

        exp_activs = np.column_stack(
            [ (Fi*vi).sum(-1) for Fi, vi in zip(F, v)])
        exp_activs = np.exp(exp_activs)

        return exp_activs / exp_activs.sum(-1)[:, None]

class EMFitMixin:
    """
    Handles MAP fitting of MLFMSA-MoE model using the EM algorithm.
    """
    def _setup_softmax(self, ifix, basis_funcs):

        assert(len(ifix) == len(basis_funcs))
        tfix = [self.ttc[i] for i in ifix]

        def wrapper(foo, ti):
            def f(x):
                return foo(x, ti)
            return f

        # wrap all the functions to demove the denese on fi
        centered_basis_funcs = []
        for bf, ti in zip(basis_funcs, tfix):
            new_bf = []
            for func in bf:
                new_bf.append(wrapper(func, ti))
            centered_basis_funcs.append(new_bf)

        self.softmax_activs = Softmax(centered_basis_funcs)

    def _get_responsibilities(self, pi, g, beta, mu_ivp, alpha_prec, ifix):
        """
        Returns
        -------
            responsibilities: [r1, ..., rQ]
        """
        probs = [[]]*len(self.N_data)  # store log probs for each replicate
        for i, ifx in enumerate(ifix):
            K = self._K(g, beta, ifx)
            Lop = np.linalg.matrix_power(K, self.order)

            means = np.kron(mu_ivp[:, i, :],
                            np.ones(self.dim.N)).T
            means = Lop.dot(means)
            
            for q, replicate in enumerate(self.y_train_):
                # subsample mean using data inds
                m_i_q = means[:, q].reshape(self.dim.K, self.dim.N).T
                m_i_q = m_i_q[self.data_inds[q], :]
                lp_i_q = norm.logpdf(
                    replicate.reshape(self.dim.K, self.N_data[q]).T,
                    loc=m_i_q,
                    scale=1/np.sqrt(alpha_prec))

                # sum over the dimesnions
                lp_i_q = lp_i_q.sum(-1)

                if probs[q] == []:
                    probs[q] = lp_i_q[:, None]
                else:
                    probs[q] = np.column_stack((probs[q], lp_i_q))

        probs = [np.exp(lp)*pi_p for lp, pi_p in zip(probs, pi)]
        sum_probs = [item.sum(-1) for item in probs]

        return [p / sum_p[:, None] for p, sum_p in zip(probs, sum_probs)]

    def _optim_g(self, g, beta, mu_ivp,
                 alpha_prec, responsibilities,
                 ifix, gprior):

        resp = [np.row_stack([r_q]*self.dim.K)
                for r_q in responsibilities]

        # inverse covariance of g
        Cg = block_diag(*[gp.kernel(self.ttc[:, None])
                          for gp in self.latentforces])
        Cg[np.diag_indices_from(Cg)] += 1e-5
        Lg = np.linalg.cholesky(Cg)

        # going to optimise mu_ivp too
        # shape (N_replicates, N_ifix, ...) they are all independent
        ones = np.ones(self.dim.N)

        # Useful dimensions
        N_repl = len(self.y_train_)
        N_mixcomp = len(ifix)        
        
        inds = [np.concatenate([self.data_inds[m] + k*self.dim.N
                                for k in range(self.dim.K)])
                for m in range(len(self.y_train_))]

        def _g_objfunc(arg):
            g = arg[:self.dim.N*self.dim.R]
            #_mu_ivp = arg[self.dim.N*self.dim.R:]
            #mu_ivp = _mu_ivp.reshape((N_repl, N_mixcomp, self.dim.K))

            log_lik, log_lik_g_grad = (0, np.zeros(g.size))
            log_lik_mu_ivp_grad = np.empty((mu_ivp.shape))

            for i, ifx in enumerate(ifix):
                # sum over the mixture components first so we don't have to
                # recalculate the gradient of the matrix operator
                Ki, Ki_grad = self._K(g, beta, ifx, eval_gradient=True)

                # ivp_means is shape [N_replic., N_ifix, self.dim.K]
                mu_ivp_ifx = mu_ivp[:, i, :]
                # means will be shape (NK, N_replicates)
                means = np.kron(mu_ivp[:, i, :],
                                np.ones(self.dim.N)).T

                # raise Ki to the correct order and get its gradient

                # is this redudnant?
                #Li = np.linalg.matrix_power(Ki, self.order)
                
                Lim, Lim_grad = Mpower_action_wgrad(Ki, Ki_grad, means, self.order)

                # sum over the replicates
                for q, replicate in enumerate(self.y_train_):
                    # sub sample the means to match up with the time
                    # indices of replicate q
                    # note data_inds[q] is an np.array
                    inds_q = inds[q]
                    Lim_q = Lim[inds_q, q]

                    eta = Lim_q - replicate
                    r_q_i = resp[q][:, i]  # responsibilities for the qth replicate
                                           # of the ith mixture component
                    Reta = eta * r_q_i     # weight eta by the responsibilites

                    # update the log-likelihood and gradient
                    log_lik += -0.5 * (eta * Reta).sum() * alpha_prec
                    log_lik_g_grad -= alpha_prec * np.array(
                        [np.sum(Reta * dL[inds_q, q]) for dL in Lim_grad])

                    # gradient with respect to the ivp mean
                    """
                    ei = np.zeros(self.dim.K)
                    for k_dim in range(self.dim.K):
                        ei[k_dim] = 1.
                        deta = Li.dot(np.kron(ei, ones))[inds_q]
                        log_lik_mu_ivp_grad[q, i, k_dim] = \
                                               - alpha_prec * deta.dot(Reta)
                        ei[k_dim] = 0.
                    """

                # add the contribution from the prior
                alpha = cho_solve((Lg, True), g)
                log_lik -= .5 * g.dot(alpha)
                log_lik_g_grad -= alpha

            grad = np.concatenate((log_lik_g_grad,))
            #log_lik_mu_ivp_grad.flatten()))
            return -log_lik, -grad

        # flatten _mu_ivp (flattens along rows
        _mu_ivp = mu_ivp.reshape(N_repl*N_mixcomp*self.dim.K)

        x0 = np.concatenate((g,))# _mu_ivp))

        # optimize
        res = minimize(_g_objfunc, x0, jac=True, tol=1e-3)
        # unpack the results
        g = res.x[:self.dim.R*self.dim.N]
        #mu_ivp = res.x[self.dim.R*self.dim.N:]
        #mu_ivp = mu_ivp.reshape(N_repl, N_mixcomp, self.dim.K)

        return g, mu_ivp

    def _optim_pi_par(self, v, responsibilities, vprior=None):
        """
        EM optimisation of the parameters of the softmax activations.
        """

        def _v_objfunc(v):
            val, grad = (0, np.zeros(len(v)))

            for q, x_q in enumerate(self.x_train_):

                pi = self.softmax_activs(v, x_q[:, None])

                val += np.sum(responsibilities[q] * np.log(pi))

                # calculate the gradient
                F = [np.column_stack((f(x_q[:, None])
                                      for f in bfuncs))
                     for bfuncs in self.softmax_activs.basis_funcs]
                
                _grad = []
                for i, Fi in enumerate(F):
                    d_i = responsibilities[q][:, i][:, None] * Fi
                    d_i -= Fi * pi[:, i][:, None]
                    _grad.append(d_i.sum(0))
                _grad = np.concatenate(_grad)
                grad += np.array(_grad)

            if vprior is not None:
                lp, lp_grad = vprior.logpd(v, eval_gradient=True)
                val += lp
                grad += lp_grad

            return -val, -grad

        
        res = minimize(_v_objfunc, x0=v, jac=True)
        return res.x

    def _initialse_Z(self):
        from scipy.interpolate import interp1d
        zz = []
        for t, y in zip(self.x_train_, self.y_train_):
            y = y.reshape(self.dim.K, t.size).T
            u = [interp1d(t, yk, kind='cubic', fill_value='extrapolate')
                 for yk in y.T]
            z = np.concatenate([uk(self.ttc) for uk in u])
            zz.append(z)

        # construct the statistics sum[zzT] sum[z_i z_{i+1}T]
        S0 = [self.order*np.outer(z, z) for z in zz]
        S1 = [self.order*np.outer(z, z) for z in zz]
        return S0, S1

    def _update_z_dist2(self, g, beta, ifx, lam, alf, mu_ivp):
        gp = self.latentforces[0]
        Cz = [gp.kernel(self.ttc[:, None])]*self.dim.K
        Lz = []
        for c in Cz:
            c[np.diag_indices_from(c)] += 1e-5
            Lz.append(np.linalg.cholesky(c))

        Cz_inv = block_diag(*[cho_solve((L, True),
                                        np.eye(L.shape[0]))
                              for L in Lz])

        K = self._K(g, beta, ifx)

        # parameters for the LDS update
        q = 0
        Sigma = np.eye(self.N_data[q]*self.dim.K) / alf

        y = self.y_train_[q]

        Gamma_inv = np.eye(self.dim.N*self.dim.K) * alf # + Cz_inv / lam


        A = K #Gamma.dot(K)
        C = np.zeros((self.N_data[q]*self.dim.K,
                      self.dim.N*self.dim.K))

        inds = self.data_inds[q]
        inds = np.concatenate([inds + self.dim.N*k
                               for k in range(self.dim.K)])
        for i in range(C.shape[0]):
            C[i, inds[i]] += 1

        Cz = block_diag(*Cz)
        Sigma = 0.01*C.dot(Cz.dot(C.T))
        Gamma = np.eye(self.dim.N*self.dim.K) / alf
        #Gamma = 0.01*np.linalg.inv(Cz_inv) #np.linalg.inv(Gamma_inv)        


        u1 = np.kron(mu_ivp[0, 0, :], np.ones(self.dim.N))
        u1 = np.zeros(self.dim.N*self.dim.K)
        V1 = np.ones((self.dim.N*self.dim.K,self.dim.N*self.dim.K))
        V1 = 100.*np.eye(V1.shape[0])

        P1 = A.dot(V1.dot(A.T)) + Gamma
        K2 = P1.dot(C.T.dot(np.linalg.inv(C.dot(P1.dot(C.T)) + Sigma)))
        u2 = A.dot(u1) + K2.dot(y - C.dot(A.dot(u1)))
        V2 = (np.eye(K2.shape[0]) - K2.dot(C)).dot(P1)

        J1 = V1.dot(A.T.dot(np.linalg.inv(P1)))

        u1h = u1 + J1.dot(u2 - A.dot(u1))
        V1h = V1 + J1.dot(V2 - P1).dot(J1.T)

        means = (u1h, u2)
        covs = (V1h, V2)
        pwcovs = (J1.dot(V2), )
        return means, covs, pwcovs
                                

    def _update_z_dist(self, g, beta, ifx, lam=1000):

        # prior inv. covariance matrix for z
        gp = self.latentforces[0]
        Cz = [gp.kernel(self.ttc[:, None])]*self.dim.K
        Lz = []
        for c in Cz:
            c[np.diag_indices_from(c)] += 1e-5
            Lz.append(np.linalg.cholesky(c))

        Cz_inv = block_diag(*[cho_solve((L, True),
                                        np.eye(L.shape[0]))
                              for L in Lz])

        # construct the data indices transform
        q = 0
        inds = self.data_inds[q]
        inds = np.concatenate([inds + self.dim.N*k
                               for k in range(self.dim.K)])

        C = np.zeros((self.N_data[q]*self.dim.K,
                      self.dim.N*self.dim.K))

        for i in range(C.shape[0]):
            C[i, inds[i]] += 1


        K = self._K(g, beta, ifx)
        CK = C.dot(K)

        A = C.dot(K)
        
        invcov = Cz_inv + lam * A.T.dot(A)
        premean = A.T.dot(self.y_train_[q]) * lam

        #dot(K)).T.dot(self.y_train_[q])
        cov = np.linalg.inv(invcov)
        mean = cov.dot(premean) 

        return mean, cov, K.dot(mean)

    def _update_g(self, Ez, EzzT, beta, ifx, lam):
        q = 0
        inds = self.data_inds[q]
        inds = np.concatenate([inds + self.dim.N*k
                               for k in range(self.dim.K)])
        C = np.zeros((self.N_data[q]*self.dim.K,
                      self.dim.N*self.dim.K))

        for i in range(C.shape[0]):
            C[i, inds[i]] += 1
        CTC = C.T.dot(C)

        Cg = [gp.kernel(self.ttc[:, None])
              for gp in self.latentforces]
        for c in Cg:
            c[np.diag_indices_from(c)] += 1e-5
        Lg = [np.linalg.cholesky(c) for c in Cg]

        invcov = block_diag(*[
            cho_solve((L, True), np.eye(self.dim.N*self.dim.R))
            for L in Lg])

        V, v = self._vecK_aff_rep(beta, ifx)
        S0 = np.kron(EzzT, CTC)
        invcov += lam * V.T.dot(S0.dot(V))
        premean = np.outer(C.T.dot(self.y_train_[q]), Ez).T.ravel()
        premean -= v.dot(S0)
        premean = lam * premean.dot(V)

        return np.linalg.solve(invcov, premean)

        
    def _EM_g_update(self, S0, S1, beta, ifx):
        lam = 1e5
        # inverse covariance of g
        Cg = [gp.kernel(self.ttc[:, None])
              for gp in self.latentforces]
        for c in Cg:
            c[np.diag_indices_from(c)] += 1e-5
        Lg = [np.linalg.cholesky(c) for c in Cg]

        invcov = block_diag(*[
            cho_solve((L, True), np.eye(self.dim.N*self.dim.R))
            for L in Lg])

        S_x_I = np.kron(S0, np.eye(self.dim.N*self.dim.K))

        # prior inv covariance
        V, v = self._vecK_aff_rep(beta, ifx)
        invcov += lam*V.T.dot(S_x_I.dot(V))

        # 'pre-mean'
        premean = S1.T.ravel() - v.dot(S_x_I)
        premean = lam*premean.dot(V)

        return np.linalg.solve(invcov, premean)


    def _Kalman_g_update(self, g, beta, mu_ivp, ifix, alpha_prec):
        for i, ifx in enumerate(ifix):

            lam = 1e-5
            Gamma = lam*np.diag(np.ones(self.dim.N*self.dim.K))
            Sigma = np.diag(np.ones(self.dim.N*self.dim.K) / alpha_prec )

            Ki = self._K(g, beta, ifx)
            mu_ivp_ifx = mu_ivp[:, i, :]

            mu0 = np.kron(mu_ivp[:, i, :],
                          np.ones(self.dim.N)).T

            muf, Vf = ([], [])  # mean and covar. for the formward sweep of
                                # the Kalman filter

            #P0 = Gamma#1e-4*np.eye(self.dim.N*self.dim.K)
            P0 = lam*np.ones((self.dim.N*self.dim.K,
                            self.dim.N*self.dim.K))
            Pchol = []

            ## Forwrad sweep
            for nt in range(self.order-1):
                if nt == 0:
                    muf.append(mu0)
                    Vf.append(P0)
                else:
                    muf.append(Ki.dot(muf[-1]))
                    Vf.append(Ki.dot(Ki.dot(Vf[-1]).T))

                P = Ki.dot(Ki.dot(Vf[-1]).T) + Gamma
                P[np.diag_indices_from(P)] += 1e-6
                Pchol.append(np.linalg.cholesky(P))

            # Now update final term in the forward sweep which contains
            # the data
            _m = Ki.dot(muf[-1])
            mfinal = _m.copy()

            for q, inds in enumerate(self.data_inds):

                inds = np.concatenate([inds + self.dim.N*k
                                       for k in range(self.dim.K)])
                C = np.zeros((self.dim.N*self.dim.K, self.N_data[q]*self.dim.K))
                for i in range(C.shape[1]):
                    C[inds[i], i] += 1
                C=C.T

                Sigma = np.diag(np.ones(self.N_data[q]*self.dim.K) / alpha_prec )

                gp = self.latentforces[0]
                #Sigma = lam*block_diag(*[gp.kernel(self.x_train_[q][:, None])
                #                         for k in range(self.dim.K)])
                
                inds_ix = np.ix_(inds, inds)
                #CPCt = P[inds_ix]  # subsample for where we have data
                CPCt = P[inds_ix]

                # shape (Naug, Ndat_q)
                Kgain = P.dot(C.T.dot(np.linalg.inv(CPCt + Sigma)))

                mfinal[:, q] += Kgain.dot(self.y_train_[q] - C.dot(_m[:, q]))

                #Vfinal = (np.eye(self.N_data[q]*self.dim.K) - \
                #          Kgain[inds, :]).dot(P[inds, :])
                Vfinal = (np.eye(Kgain.shape[0]) - Kgain.dot(C)).dot(P)

                muf.append(mfinal)
                Vf.append(Vfinal)
            # Backward sweep
            pw_covs = []
            KiT = Ki.T

            for nt in range(self.order-1):

                Jn = Vf[-(nt+2)].dot(
                    KiT.dot(
                    cho_solve((Pchol[-(nt+1)], True),
                              np.eye(self.dim.N*self.dim.K))))

                mn = muf[-(nt+2)] + \
                     Jn.dot(muf[-(nt+1)] - Ki.dot(muf[-(nt+2)]))

                Pn = Pchol[-(nt+1)].dot(Pchol[-(nt+1)].T)
                Vn = Vf[-(nt+2)] - \
                     Jn.dot(Jn.dot(Vf[-(nt+1)] - Pn).T)
                muf[-(nt+2)] = mn
                Vf[-(nt+2)] = Vn

                pw_covs.append(Jn.dot(Vf[-(nt+1)]))

            # get the two statistics of interest
            S0 = 0.
            for V, m in zip(Vf[:-1], muf[:-1]):
                S0 += V + np.outer(m[:, 0], m[:, 0])
            S1 = 0.
            for i, pwV in enumerate(pw_covs):
                S1 += pw_covs[0] + np.outer(muf[i][:, 0], muf[i+1][:, 0])

            return S0, S1, muf, Vfinal

            # repr. of Ki = v0 + V.dot(g)
            V, v0 = self._vecK_aff_rep(beta, ifx)

            S0 = np.kron(S0, np.eye(self.dim.N*self.dim.K))

            expr1 = V.T.dot(S0.dot(V))
            expr2 = S1.T.ravel() - v0[None, :].dot(S0)
            expr2 = expr2.dot(V)

            # inverse covariance of g
            Cg = [gp.kernel(self.ttc[:, None])
                  for gp in self.latentforces]
            for c in Cg:
                c[np.diag_indices_from(c)] += 1e-5
            Lg = [np.linalg.cholesky(c) for c in Cg]

            inv_cov = block_diag(*[
                cho_solve((L, True), np.eye(self.dim.N*self.dim.R))
                for L in Lg])
            #inv_cov *= lam
            #inv_cov += expr1
            c = np.linalg.inv(inv_cov)


            mean = np.linalg.solve(inv_cov, expr2[0, :])
            
            z0 = muf[-2][:, 0]
            z1 = muf[-1][:, 0]

            ExxT = Vf[-2] + np.outer(z0, z0)

            A = np.kron(z0[None, :], np.eye(self.dim.N*self.dim.K))
            A = A.dot(V)
            b = np.kron(z0[None, :], np.eye(self.dim.N*self.dim.K)).dot(v0)

            #A += np.row_stack([inv_cov]*self.dim.K)
            
            mean = np.linalg.lstsq(A, z1 - b, rcond=None)
            rank = mean[2]
            mean = mean[0]

            ic = V.T.dot(np.kron(np.outer(z0, z0),
                                 np.eye(self.dim.N*self.dim.K))).dot(V)
            ic += inv_cov * lam
            pm = np.kron(z0[:, None], np.eye(self.dim.N*self.dim.K)).dot(z1[:, None])
            pm -= np.kron(np.outer(z0, z0), np.eye(self.dim.N*self.dim.K)).dot(v0[:, None])
            pm = V.T.dot(pm)
#            pm /= lam

            mean = np.linalg.solve(ic, pm)
            
            return muf[-1], mean



class BaseMLFMSA(BaseMLFM):

    is_tt_aug = False  # allows the possibility to augment the time vector

    def __init__(self, *args, order=1, **kwargs):
        super(BaseMLFMSA, self).__init__(*args, **kwargs)

        if isinstance(order, int) and order > 0:
            self.order = order

        else:
            raise ValueError("The approximation error must be a"
                             " positive integer.")

    def _setup_times(self, tt, tt_aug=None,
                     h=None, multi_output=False):

        if not multi_output:
                tt = [tt, ]

        # numpy arrayify time vectors
        tt = [np.asarray(item) for item in tt]
        
        if True:# h is not None:
            # h is checked first
            ttc, data_inds = handle_time_inds(tt, h)
            self.x_train_ = tt
            self.ttc = ttc
            self.data_inds = data_inds
            self.is_tt_aug = True

        elif tt_aug is not None:
            # User has supplied the time points at which they want
            # to estimate the latent forces

            if multi_output:
                # add tt_aug to the times list
                tt = [tt_aug] + [t for t in tt]
                ttc, data_inds = handle_time_inds(tt, h)
                self.x_train_ = tt[1:]
                self.ttc = ttc
                self.data_inds = data_inds[1:]
                self.is_tt_aug = True

        else:
            # neither h or tt_aug is supplied
            self.x_train_ = tt
            self.ttc = ttc
            self.is_tt_aug = None

        # store the dimension variables
        _, K, R, D = self.dim
        self.dim = Dimensions(self.ttc.size, K, R, D)
        self.N_data = tuple(item.size for item in self.x_train_)

    def _get_weight_matrix(self, tt, i):
        """
        Weight matrix for the trapezoidal rule.
        """

        def _get_weights(tt):
            tt = np.asarray(tt)
            W = np.zeros((tt.size, tt.size))
            h = np.diff(tt)
            for i in range(len(tt)):
                W[i, :i] += .5 * h[:i]
                W[i, 1:i+1] += .5 * h[:i]
            return W

        ttb = tt[:i+1]
        ttf = tt[i:]

        Wb = _get_weights([t for t in reversed(ttb)])
        Wb = np.flip(Wb, (0, 1))
        Wf = _get_weights(ttf)

        W = np.zeros((len(tt), len(tt)))
        W[:i+1, :i+1] += Wb
        W[i:, i:] += Wf

        return W

    def _K(self, g, beta, ifix, eval_gradient=False, eval_b_gradient=False):
        """
        Return the discretisation of the integral opearator in Picard iteration.
        """
        W = self._get_weight_matrix(self.ttc, ifix)
        g = g.reshape((self.dim.R, self.dim.N)).T
        g = np.column_stack((np.ones(self.dim.N), g))
        
        # struct matrices A_r = sum_r g_rn * sum_d brd * Ld
        struct_mats = np.array([sum(brd*Ld
                                    for brd, Ld in zip(br, self.basis_mats))
                                for br in beta])
        At = struct_mats[None, ...] * g[..., None, None]
        At = At.sum(1)  # sum over R
        
        eifx = np.eye(1, self.dim.N, ifix)
        
        K = sum(np.kron(At[i, ...],
                        np.kron(W[:, i][:, None], np.eye(1, self.dim.N, i)))
                for i in range(self.dim.N))

        I = np.eye(self.dim.K)
        
        K += sum(np.kron(I, np.kron(np.eye(1, self.dim.N, i).T, eifx))
                 for i in range(self.dim.N))

        if eval_gradient:
            K_grad = np.array([np.kron(struct_mats[r+1],
                                       np.kron(W[:, i][:, None],
                                               np.eye(1, self.dim.N, i)))
                                   for r in range(self.dim.R)
                               for i in range(self.dim.N)])
            if eval_b_gradient:
                # This is lazy
                dB = self._dKdB(g, beta, ifix)
                return K, K_grad, dB

            return K, K_grad
        else:
            return K

    def _dKdB(self, g, beta, ifix):
        """
        Rewriting of this function to reduce redundant computations
        """
        W = self._get_weight_matrix(self.ttc, ifix)
        

        # for computational efficiency we need to do
        # any scalar multiplications *before* any
        # kronecker products
        GW = np.stack((gn[:, None] * W[:, n][None, :]
                       for n, gn in enumerate(g)))

        dKdB = []
        for r in range(self.dim.R+1):
            for d in range(self.dim.D):
                dKdB.append(np.kron(self.basis_mats[d],
                                    GW[:, r, :].T))

        return dKdB



    def _Kx_power(self, x,
                  g, beta, power, ifx,
                  eval_g_grad=False, eval_b_grad=False):
        pass
        #if eval_g_grad and eval_b_grad:
        #    K, dKdg, dKdb = self._K(...)


    

    def _Kx(self, G, beta, x, ifx):
        """
        Returns the action of K on a vector x
        """
        xshape = x.shape         # K acts on the final dimension of x
        naxes = len(x.shape) - 1 # number of axes to ignore
        if x.shape[-1] != self.dim.N*self.dim.K:
            raise ValueError("x[..., -1] must be of size {}{}".format(self.dim.K, self.dim.N))

        X = x.reshape(xshape[:-1] + (self.dim.K, self.dim.N))
        X = X.transpose(tuple(range(naxes)) + (naxes+1, naxes))

        # XL[..., N, D, ...] = X[..., :, :].dot(L[d, ...])
        XL = np.einsum('...jk,...dlk->...djl', X, self.basis_mats)

        # get the quadrature weight matrix
        W = self._get_weight_matrix(self.ttc, ifx)

        # WXL shape (..., D, N, N, K)
        # WXL[..., d, n, ...] gives the outer product of the nth
        # col of W with the the nth row of XL[d, ...]
        WXL = (W.T)[..., None] * XL[..., None, :]
        
        # transpose and then flatten so that last component of array
        # is a vectorisation
        nshape = len(WXL.shape)
        perm = tuple(range(nshape-2)) + (nshape-1, nshape-2)
        WXL = WXL.transpose(perm)
        WXL = WXL.reshape(WXL.shape[:-2] + (self.dim.K*self.dim.N, ))

        # Gradient of action with respect to latent forces
        #  - multiply be beta and then sum over d=1,..., D
        #dG = beta[..., None, None] * WXL[..., None, :, :, :]
        #dG = dG.sum(2)

        # Gradient of action with respect to beta
        #  - multiply by G and then sum over n=1,...,N

        # pad G with col of ones
        Gt = G.T

        
        dB = Gt[:, None, :, None] * WXL[..., None, :, :, :]
        dB = dB.sum(3)

        # typically dB is smaller so go for
        Kx = beta[..., None] * dB[..., :, :, :]
        Kx = Kx.sum((1, 2))
        
        # Get the fixed value 
        fixedval = x[..., ifx::self.dim.N]
        xfix = np.kron(fixedval, np.ones(self.dim.N))

        Kx += xfix

        return Kx#, dG, dB

    def _Kx_power(self, g, beta, ifx, x, power):
        # stored a single copy of K
        K = self._K(g, beta, ifx)
        G = np.column_stack((np.ones(self.dim.N),
                             g.reshape(self.dim.R, self.dim.N).T))

        res = x

        gradshape = x.shape[:-1] + (self.dim.R, self.dim.N, x.shape[-1])
        grad = np.empty(gradshape)

        for m in range(power):
            # get K^m x, dK K^{m-1}x
            #res, dKm, _ = self._Kx(G, beta, res, ifx)
            res = self._Kx(G, beta, res, ifx)

            # get K.dot( d [K^{m-1}] )
            #KdKmprev = np.einsum('ij,...nrj->...nri', K, grad)

            # update grad
            #grad = dKm[..., 1:, :, :] + KdKmprev

        return res, grad

    def _vecK_aff_rep(self, beta, ifix, eval_gradient=False):

        struct_mats = np.array([sum(brd*Ld
                                    for brd, Ld in zip(br, self.basis_mats))
                                for br in beta])

        W = self._get_weight_matrix(self.ttc, ifix)

        I = np.eye(self.dim.K)

        NK = self.dim.N*self.dim.K
        K = self.dim.K
        cols = []
        for r in range(self.dim.R):
            for n in range(self.dim.N):
                expr = np.kron(struct_mats[r+1],
                               np.kron(W[:, n][:, None],
                                       np.eye(1, self.dim.N, n)))
                cols.append(expr.T.ravel())
        V = np.column_stack(cols)
        v0 = self._K(np.zeros(self.dim.N*self.dim.R), beta, ifix).T.ravel()
        return V, v0
        
    def bar(self, g, beta, mu_ivp,
            alpha_prec, ifix):
        x = np.kron(mu_ivp[:, 0, :],
                    np.ones(self.dim.N)).T
        x = x[:, 0]
        K = self._K(g, beta, 0)

        x1 = K.dot(x)
        x2 = K.dot(x1)
        x3 = K.dot(x2)



        ########
        ifx = 0
        z0 = x2
        z1 = x3
        lam = 1e5

        # repr. of Ki = v0 + V.dot(g)
        V, v0 = self._vecK_aff_rep(beta, ifx)        


        A = np.kron(z0[None, :], np.eye(self.dim.N*self.dim.K))
        A = A.dot(V)
        b = np.kron(z0[None, :], np.eye(self.dim.N*self.dim.K)).dot(v0)

        #A += np.row_stack([inv_cov]*self.dim.K)
            
        mean = np.linalg.lstsq(A, z1 - b, rcond=None)
        rank = mean[2]
        mean = mean[0]
        

        import matplotlib.pyplot as plt
        fig, ax = plt.subplots()
        ax.plot(self.ttc, g, 'C0-')
        ax.plot(self.ttc, mean, 'r+')
        ax.set_ylim((-2, 2))
        plt.show()

    def fooz(self, g, beta, mu_ivp,
             alpha_prec, responsibilites, gtrue):
        """
        Constrcts an estimate of the distribution of Z
        """
        mean, cov, Kmean = self._update_z_dist(g, beta, 0, lam=10000)
        
        import matplotlib.pyplot as plt
        mean = mean.reshape(self.dim.K, self.dim.N).T
        Kmean = Kmean.reshape(self.dim.K, self.dim.N).T

        _m, _covs, _pwcovs = self._update_z_dist2(g, beta, 0, 1e4, 1e4, mu_ivp)

        _mean = _m[0].reshape(self.dim.K, self.dim.N).T
        _Kmean = _m[1].reshape(self.dim.K, self.dim.N).T

        S0 = np.outer(_m[0], _m[0]) + _covs[0]
        S1 = np.outer(_m[0], _m[1]) + _pwcovs[0]

        """
        fig, ax = plt.subplots()
        
        ax.plot(self.ttc, mean, 'r-')
        ax.plot(self.ttc,
                Kmean, 'o')
        ax.plot(self.x_train_[0],
                self.y_train_[0].reshape(self.dim.K, self.N_data[0]).T, 'C0--', alpha=0.4)
        """

        fig, ax = plt.subplots()
        ax.plot(self.ttc, _mean, 'b-')
        ax.plot(self.ttc, _Kmean, 'r-')
        ax.plot(self.x_train_[0],
                self.y_train_[0].reshape(self.dim.K, self.N_data[0]).T, 'C0--', alpha=0.4)

        ifx = 0
        #z0 = mean.T.ravel()
        z0 = _m[0]
        z1 = _m[1]
        #z1 = Kmean.T.ravel()
        lam = 1e5

        # repr. of Ki = v0 + V.dot(g)
        V, v0 = self._vecK_aff_rep(beta, ifx)        

        A = np.kron(z0[None, :], np.eye(self.dim.N*self.dim.K))
        A = A.dot(V)
        b = np.kron(z0[None, :], np.eye(self.dim.N*self.dim.K)).dot(v0)

        #S0 = np.outer(z0, z0) + cov
        #S1 = np.outer(z0, z1) + 1e-5*np.eye(z1.size)
        _A = V.T.dot(np.kron(S0, np.eye(self.dim.N*self.dim.K)).dot(V))
        
        Cg = [gp.kernel(self.ttc[:, None])
              for gp in self.latentforces]
        for c in Cg:
            c[np.diag_indices_from(c)] += 1e-5
        Lg = [np.linalg.cholesky(c) for c in Cg]

        invcov = block_diag(*[
            cho_solve((L, True), np.eye(self.dim.N*self.dim.R))
            for L in Lg])
        
        _A += invcov / lam
        _b = (S1.ravel() - v0.dot(np.kron(S0, np.eye(self.dim.N*self.dim.K)))).dot(V)

        #A += np.row_stack([inv_cov]*self.dim.K)
        _mean = np.linalg.lstsq(_A, _b, rcond=None)
            
        mean = np.linalg.lstsq(A, z1 - b, rcond=None)

        rank = mean[2]
        mean = mean[0]
        
        #mean = np.linalg.solve(ic, pm)

        fig, ax = plt.subplots()
        ax.plot(self.ttc, _mean[0], '-.')
        #ax.plot(self.ttc, mean, '--')
        #ax.plot(self.ttc, g, '+')
        ax.plot(self.ttc, gtrue, 'k-', alpha=0.4)
        plt.show()
        assert(False)
        return _mean[0]
        
    def foo(self, g, beta, mu_ivp,
            alpha_prec, responsibilites,
            ifix):

        N_repl = len(self.y_train_)

        resp = [np.row_stack([r_q]*self.dim.K)
                for r_q in responsibilites]

        inds = [np.concatenate([self.data_inds[m] + k*self.dim.N
                                for k in range(self.dim.K)])
                for m in range(N_repl)]

        ones = np.ones(self.dim.N)  # precompute a useful vector of ones

        def objfunc(g):
            ifx = 0

            Ki, Ki_grad = self._K(g, beta, ifx, eval_gradient=True)
            return Ki_grad
            #means = np.kron(mu_ivp[:, 0, :],
            #                ones).T
            #L, Lgrad = Mpower_action_wgrad(Ki, Ki_grad, means, self.order)
            #
            #return Lgrad

        def bar(g):
            x = np.kron(mu_ivp[:, 0, :],
                        ones).T
            # X[M_repl, K, N]  X[n, k, :]  kth dimension component of X
            X = x.T.reshape((N_repl, self.dim.K, self.dim.N))
            X = X.transpose((0, 2, 1))  # transpose the last two axes of X
            Arr = np.dstack([sum(brd*Ld
                                 for brd, Ld in zip(br, self.basis_mats)).T
                             for br in beta])

            # XA[n, ..., r] = X[n, ...].dot(A[...r])
            XA = np.einsum('ijk,klr->ijlr', X, Arr)


            W = self._get_weight_matrix(self.ttc, 0)
            # Want to store the outer product of the nth column of W
            # with the nth row X[, n, ]
            n = 3
            d = 1
            # broadcasting
            #     W        XA
            #   (N, N)   (M, N, K, R)
            #
            # WXA[m, n, :, k, r] = outer prod of the nth col. of W with
            # the mth replicate of X
            WXA = (W.T)[None, :, :, None, None] * XA[:, :, None, :, :]

            # finally we want to transpose and then reshape so that X is
            # vectorised
            WXA = WXA.transpose((0, 1, 4, 3, 2))
            return WXA.reshape((N_repl, self.dim.N, self.dim.R+1, self.dim.N*self.dim.K))

        def bar2(g, x):
            xshape = x.shape
            naxes = len(x.shape) - 1
            X = x.reshape(xshape[:-1] + (self.dim.K, self.dim.N))
            X = X.transpose(tuple(range(naxes)) + (naxes+1, naxes))

            Arr = np.stack([sum(brd*Ld
                                for brd, Ld in zip(br, self.basis_mats))
                            for br in beta])
            XA = np.einsum('...jk,...dlk->...djl', X, Arr)
            W = self._get_weight_matrix(self.ttc, 0)
            WXA = (W.T)[:, None, :, None] * XA[..., None, :, :, :]

            nshape = len(WXA.shape)
            perm = tuple(range(nshape-2)) + (nshape-1, nshape-2)
            WXA = WXA.transpose(perm)

            # Transpose and flatten the last two axes
            WXA = WXA.reshape(WXA.shape[:-2] + (self.dim.K*self.dim.N, ))
            return WXA

        def bar3(g, x):
            """
            Returns K[g, beta].dot(x[..., ]) and optionaly its gradient

            g_grad : shape (..., N, R, NK)
                g_grad[..., n, r, :] gives the gradient of K.dot(x[..., :])
                w.r.t g_{r, n}

            b_grad : shape (..., R, D, NK)
                b_grad[..., r, d, :] gives the gradient of K.dot(x[..., :])
                w.r.t. to beta_{r, d}
            """
            xshape = x.shape
            naxes = len(x.shape) - 1

            X = x.reshape(xshape[:-1] + (self.dim.K, self.dim.N))
            X = X.transpose(tuple(range(naxes)) + (naxes+1, naxes))

            XL = np.einsum('...jk,...dlk->...djl', X, self.basis_mats)
            W = self._get_weight_matrix(self.ttc, 0)
            WXL = (W.T)[:, None, :, None] * XL[..., None, :, :, :]

            nshape = len(WXL.shape)
            perm = tuple(range(nshape-2)) + (nshape-1, nshape-2)
            WXL = WXL.transpose(perm)

            # Transpose and flatten the last two axes
            WXL = WXL.reshape(WXL.shape[:-2] + (self.dim.K*self.dim.N, ))

            # multiply by beta along D and make a new axis for R


            # To get the gradient of the action with respect to
            # the latent forces multiply be beta and then sum over
            # d = 1,..,D
            WXA = beta[..., None] * WXL[..., None, :, :]            
            WXA = WXA.sum(3)

            # to get the gradient of the action wrt to
            # beta, multiply by G and then sum over n
            dB = G[..., None, None] * WXL[..., None, :, :]
            dB = dB.sum(1)
            return WXA, dB
            

        D1 = objfunc(g)
        x = np.kron(mu_ivp[:, 0, :], ones).T

        m = 1
        n = 4
        val1 = D1[n, ...].dot(x[:, m])

        D2 = bar(g)
        val2 = D2[m, n, 1, :]

        K = self._K(g, beta, 0)
        _Kx = K.dot(x)

        ifx = 0
        # rip out the fixed initial condition
        fixedval = x[ifx::self.dim.N, :].T
        xfix = np.kron(fixedval, ones).T

        G = np.column_stack((ones, g.reshape(self.dim.R, self.dim.N).T))
        Kx = (G[None, ..., None]) * D2
        Kx = Kx.sum((1, 2))
        Kx += xfix.T

        y = x.T
        # add a dummy extra axis
        
        
        D3 = bar2(g, y)
        D4, dB = bar3(g, y)

        def _foo(b):
            K = self._K(g, b, 0)
            return K.dot(y.T).T

        res = _foo(beta)
        r = 1
        d = 1
        eps = 1e-6
        betap = beta.copy()
        betap[r, d] += eps
        f = _foo(beta)
        fp = _foo(betap)
        dnum = (fp-f)/eps


        z = np.random.normal(size=y.size).reshape(y.shape)


        K, dK = self._K(g, beta, 0, True)


        power = 20
        import time
        t0 = time.time()
        pow2, _ = Mpower_action_wgrad(K, dK, y.T, 2)
        t1 = time.time()
        print("Old method:", t1 - t0)

        t0 = time.time()
        pow1, powergrad = self._Kx_power(g, beta, 0, y, power)
        t1 = time.time()
        print("New method:", t1 - t0)



class MLFMSA(EMFitMixin, BaseMLFMSA):

    def __init__(self, *args, **kwargs):
        super(MLFMSA, self).__init__(*args, **kwargs)


    def fit(self, times, Y, multi_output=False, **kwargs):

        self._setup_times(times,
                          kwargs.pop('h', None),
                          multi_output=multi_output)


        
        


    def _fit_init(self, is_fixed_vars, **kwargs):
        """
        Handles initalisation of the optimisation
        """
        init_strategies = {
            'g': lambda : np.zeros(self.dim.N*self.dim.R),
            'beta': lambda : np.row_stack((np.zeros(self.dim.D),
                                           np.eye(self.dim.R, self.dim.D))).ravel(),
            'v': lambda : -np.ones(sum(self.softmax_activs.param_shape))
            }

        var_names = ['g', 'beta', 'v']
        full_init = [kwargs.pop("".join((vn, '0')),
                                init_strategies[vn]())
                     for vn in var_names]
        full_init_shape = [item.size for item in full_init]

        free_vars, fixed_vars = ([], [])
        for item, boolean in zip(full_init, is_fixed_vars):
            if boolean:
                fixed_vars.append(item)
            else:
                free_vars.append(item)
        free_vars_shape = [item.size for item in free_vars]

        return np.concatenate(free_vars), free_vars_shape, fixed_vars




import numpy as np
from sklearn.base import clone
from sklearn.gaussian_process import GaussianProcessRegressor
import sklearn.gaussian_process.kernels as sklearn_kernels
from scipy.special import wofz
from collections import namedtuple

_known_transform = {'log': lambda x: np.log(x),
                    'Id': lambda x: x}
_known_inv_transform = {'log': lambda x: np.exp(x),
                        'Id': lambda x: x}

class Hyperparameter(namedtuple('Hyperparameter',
                                ('name', 'value_type', 'bounds',
                                 'n_elements', 'fixed',
                                 'transform', 'inverse_transform'))):
    """A kernel hyperparamers's specification in form of a namedtuple.

    .. note
    An extension of the sklearn hyperparameter to allow the
    specifaction of the parameter transform
    """
    __slots__ = ()

    def __new__(cls, name, value_type, bounds, n_elements=1, fixed=None, transform='log'):
        bounds = np.atleast_2d(bounds)
        if n_elements > 1:  # vector-valued parameter
            if bounds.shape[0] == 1:
                bounds = np.repeat(bounds, n_elements, 0)
            elif bounds.shape[0] != n_elements:
                raise ValueError("Bounds on %s should have either 1 or "
                                 "%d dimensions. Given are %d"
                                 % (name, n_elements, bounds.shape[0]))
        if fixed is None:
            fixed = False
            #fixed = isinstance(bounds, six.string_types) and bounds == "fixed"

        try:
            transform, inv_transform = \
                       (_known_transform[transform], _known_inv_transform[transform])

        except KeyError:
            raise NotImplementedError("Unknown transform % s" % transform)
            
        return super(Hyperparameter, cls).__new__(
            cls, name, value_type, bounds, n_elements, fixed, transform, inv_transform)

    def __eq__(self, other):
        return (self, name == other.name and
                self.value_type == other.value_type and
                np.all(self.bounds == other.bounds) and
                self.n_elements == other.n_elements and
                self.fixed == other.fixed and
                self.transform == other.transform)

class Kernel(sklearn_kernels.Kernel):

    @property
    def theta(self):
        """ Returns the (flattened, 'transformed') non-fixed hyperparameters.
        
        Note that the transformation is specified during the initialisation of the kernel.
        """
        theta = []
        params = self.get_params()
        for hyperparameter in self.hyperparameters:
            if not hyperparameter.fixed:
                # transform the hyperparameter
                _theta = hyperparameter.transform(params[hyperparameter.name])
                theta.append(_theta)
        if len(theta) > 0:
            return np.hstack(theta)
        else:
            return np.array([])

    @theta.setter
    def theta(self, theta):
        params = self.get_params()
        i = 0
        for hyperparameter in self.hyperparameters:
            if hyperparameter.fixed:
                continue
            if hyperparameter.n_elements > 1:
                # vector-valued parameter
                par = hyperparameter.inverse_transform(theta[i:i + hyperparameter.n_elements])
                params[hyperparameter.name] = par
                i += hyperparameter.n_elements
            else:
                params[hyperparameter.name] = hyperparameter.inverse_transform(theta[i])
                i += 1

        if i != len(theta):
            raise ValueError("theta has not got the correct number of entries."
                             " Should be %d; given are %d"
                             % (i, len(theta)))

        self.set_params(**params)

    @property
    def bounds(self):
        """Returns the transformed bounds of theta.

        Returns
        -------
        bounds : array, shape (n_dims, 2)
        """
        bounds = []
        for hyperparameter in self.hyperparameters:
            if not hyperparameter.fixed:
                bounds.append(hyperparameter.transform(hyperparameter.bounds))
        if len(bounds) > 0:
            return (np.vstack(bounds))
        else:
            return np.array([])

def wofz_grad(z):
    return -2*z*wofz(z) + 1j*2/np.sqrt(np.pi)

def Upsilon(t, t2, gam, l_scale, eval_gradient=False):
    """
    Describe me without a reference?
    """
    # array axes ordering [ t , t2 , gamma, length scales]
    
    # pwd of times
    Dt = t[:, None] - t2[None, :]

    # u1 = 2exp(arg1)
    arg1 = .25*gam[:, None] **2 * l_scale[None, :] ** 2
    arg1 = arg1[None, None, ...] - \
           Dt[..., None, None] * gam[None, None, :, None]

    # u2 = exp(arg2) * wofz(1j * zrq)
    zrq = Dt[..., None, None] / l_scale - \
          .5 * gam[None, None, :, None] * l_scale  # easy b.-casting over last axis

    arg2 = - (Dt**2)[..., None, None] / l_scale**2

    # u3 = exp(arg3) * wofz(-1j * zrq0)
    zrq0 = -t2[None, :, None, None] / l_scale - \
           .5 * gam[None, None, :, None] * l_scale

    arg3 = -(t2**2)[None, :, None, None] / l_scale**2 -\
           t[:, None, None, None] * gam[None, None, :, None]

    if eval_gradient:

        # gradient of arg1 wrt gam
        darg1_g = .5*gam[:, None] * (l_scale ** 2)[None, :]
        darg1_g = darg1_g[None, None, ...] - \
                  Dt[..., None, None]

        u1 = 2 * np.exp(arg1) # for clarity.
        du1 = u1 * darg1_g

        u2 = np.exp(arg2) * wofz(1j * zrq)
        du2 = -.5*1j * wofz_grad(1j * zrq) * np.exp(arg2) * l_scale

        u3 = np.exp(arg3) * wofz(-1j*zrq0)
        du3 = np.exp(arg3) * \
              (-t[:, None, None, None] * wofz(-1j * zrq0) + \
               .5 * 1j * wofz_grad(-1j * zrq0) * l_scale)

        return u1 - u2 - u3, du1 - du2 - du3

    else:
        return 2*np.exp(arg1) - \
               np.exp(arg2)*wofz(1j*zrq) - \
               np.exp(arg3)*wofz(-1j*zrq0)

def Aicht2(t, t2, gq, gp, length_scale, eval_gradient=False):

    if eval_gradient:
        U1, U1grad = Upsilon(t2, t, gq, length_scale, True)
        # Add axis for gp
        U1 = U1[..., None, :]
        U1grad = U1grad[..., None, :]

        U2, U2grad = Upsilon(t2, np.zeros(t.size), gq, length_scale, True)
        # Add axis for gp
        U2 = U2[..., np.newaxis, :]
        U2grad = U2grad[..., np.newaxis, :]

        h = np.exp(-t[:, None]*gp[None, :])
        # - add axis for t2, gq, lr
        h = h[np.newaxis, :, np.newaxis, :, np.newaxis]
        dH = -h*t[None, :, None, None, None]

        H = U1 - h * U2
        gq_plus_gp = gq[:, None] + gp[None, :]
        # Add axis for t2, t, lr

        DH1grad = np.zeros((t2.size, t.size, gq.size, gp.size,
                            length_scale.size, gq.size),
                           dtype=np.complex128)
        DH2grad = np.zeros(DH1grad.shape,
                           dtype=np.complex128)

        for k in range(gq.size):
            DH1grad[..., k, :, :, k] += (U1grad[..., k, :, :] - 
                                         h[..., 0, :, :] * U2grad[..., k, :, :]) / \
                                         gq_plus_gp[None, None, k, :, None]

            DH1grad[..., k, :, :, k] -= H[..., k, :, :] / (gq_plus_gp[None, None, k, :, None])**2
            
            DH2grad[..., k, :, k] -= dH[..., k, :] * U2[..., 0, :] / \
                                     gq_plus_gp[None, None, :, k, None]

            DH2grad[..., k, :, k] -= H[..., k, :] / (gq_plus_gp[None, None, :, k, None])**2
                                     

        return H / gq_plus_gp[None, None, ..., None], \
               DH1grad, \
               DH2grad

    else:
        U1 = Upsilon(t2, t, gq, length_scale)[..., None, :]
        U2 = Upsilon(t2, np.zeros(t.size), gq, length_scale)[..., None, :]

        # Add axis for t2, gq, lr
        H = np.exp(-t[:, None] * gp[None, :])[None, :, None, :, None]
        H = U1 - H * U2
        
        gq_plus_gp = gq[:, None] + gp[None, :]
        # Add axis for t2, t, lr
        gq_plus_gp = gq_plus_gp[np.newaxis, np.newaxis, ..., np.newaxis]        

        return H / gq_plus_gp


class LFMorder2Kernel(Kernel):

    def __init__(self,
                 C=1.0, C_bounds=(-1e5, 1e5),
                 D=1.0, D_bounds=(-1e5, 1e5),
                 S=[1.], S_bounds=(-1e5, 1e5),
                 length_scale=1., length_scale_bounds=(1e-5, 1e5)):

        self.C = np.asarray(C)
        self.C_bounds = C_bounds
        self.S = np.asarray(S)
        self.S_bounds = S_bounds
        self.D = np.asarray(D)
        self.D_bounds = D_bounds
        self.length_scale = length_scale
        self.length_scale_bounds = length_scale_bounds

        # some dimensionality checks
        if self.C.size == self.D.size:
            self.n_outputs = self.C.size
        else:
            raise ValueError("C and D must be the same size")
        n_inputs = self.S.size // self.C.size
        if n_inputs > 0:
            self.n_inputs = n_inputs
        else:
            raise ValueError("S must be of size n_inputs * n_outputs")

        if isinstance(length_scale, (int, float)):
            self.n_latentforces = 1
        else:
            self.n_latentforces = len(self.length_scale)

    def is_stationary(self):
        """ Returns whether the kernel is stationary. """
        return False

    @property
    def hyperparameter_C(self):
        return Hyperparameter(
            "C", "numeric", self.C_bounds, self.C.size, transform='Id')

    @property
    def hyperparameter_D(self):
        return Hyperparameter(
            "D", "numeric", self.D_bounds, self.D.size, transform='Id')

    @property
    def hyperparameter_S(self):
        return Hyperparameter(
            "S", "numeric", self.S_bounds,
            self.S.size, transform='Id')

    @property
    def hyperparameter_length_scale(self):
        if isinstance(self.length_scale, (float, int)):
            return Hyperparameter("length_scale", "numeric",
                                  self.length_scale_bounds, fixed=True)
        else:
            return Hyperparameter("length_scale", "numeric",
                                  self.length_scale_bounds,
                                  len(self.length_scale), fixed=True)
                                  
            

    def diag(self, X):
        """
        Should be possible to speed this up.
        """
        return np.diag(self(X)).copy()

    def __call__(self, X, Y=None, eval_gradient=False):
        """ Return the kernel k(X, Y) and optionally its gradient.

        Parameters
        ----------
        X : array, shape (n_samples_X, 1)
            Left argument of the returned kernel k(X, Y)

        Y : array, shape (n_samples_Y, 1), (optional, default=None)
            Right argument of the returned kernel k(X, Y). If None, k(X, X)
            if evaluated instead.

        eval_gradient : bool (optional, default=False)
            Determines whether the gradient with respect to the kernel
            hyperparameter is determined. Only supported when Y is None.

        Returns
        -------
        K : array, shape (n_samples_X * n_output_dim, 1)
            Kernel k(X, Y)

        K_gradient : array (opt.), shape (n_samples_X * n_output_dim,
        n_samples_X * n_output_dim, n_dims)
            The gradient of the kernel k(X, X) with respect to the
            hyperparameter of the kernel. Only returned when eval_gradient
            is True
        """            
        X = np.atleast_2d(X)
        if Y is None:
            Y = X.copy()

        # get the parameters
        parameters = self.get_params()
        C = parameters['C']
        D = parameters['D']
        S = parameters['S']
        length_scale = parameters['length_scale']
        length_scale = np.atleast_1d(length_scale)

        # get some dimensionality parameters...
        Q = self.n_outputs
        R = self.n_latentforces

        # slightly dirty hack to make
        # this kernel have an algebra that works with
        # the other kernels
        N1 = X.shape[0]# // Q
        N2 = Y.shape[0]# // Q

        t1 = X[:N1, 0]
        t2 = Y[:N2, 0]

        # and use them to reshape the vectorised S
        S = S.reshape((R, Q)).T


        w = 0.5*np.sqrt(4*D - C**2 + 1j*0)
        a = 0.5*C

        # gamma and its conjugate
        gam = a + 1j*w
        gamc = a - 1j*w

        # Sprod[..., r] = np.outer(S[:, r], S[:, r])
        Sprod = S[:, None, :] * S[None, :, :]
        wprod = np.outer(w, w)

        if eval_gradient:
            dadc = .5

            dwdc = -.5*C/np.sqrt(4*D - C**2 + 1j*0)
            dwdd = .5/w

            dgamdc = dadc + 1j*dwdc
            dgamdd = 1j*dwdd
            dgamcdc = dadc - 1j*dwdc
            dgamcdd = -1j*dwdd

            H11, D1H11, D2H11 = \
                 Aicht2(t1, t2, gamc, gam, length_scale, True)
            H12, D1H12, D2H12 = \
                 Aicht2(t1, t2, gam, gamc, length_scale, True)
            H13, D1H13, D2H13 = \
                 Aicht2(t1, t2, gamc, gamc, length_scale, True)
            H14, D1H14, D2H14 = \
                 Aicht2(t1, t2, gam, gam, length_scale, True)

            H21, D1H21, D2H21 = \
                 Aicht2(t2, t1, gam, gamc, length_scale, True)
            H22, D1H22, D2H22 = \
                 Aicht2(t2, t1, gamc, gam, length_scale, True)
            H23, D1H23, D2H23 = \
                 Aicht2(t2, t1, gamc, gamc, length_scale, True)
            H24, D1H24, D2H24 = \
                 Aicht2(t2, t1, gam, gam, length_scale, True)


            D_grad1 = np.zeros((N2, N1, Q, Q, R, Q), dtype=np.complex128)
            D_grad2 = np.zeros((N1, N2, Q, Q, R, Q), dtype=np.complex128)

            grad1 = D1H11 * dgamcdc + D2H11 * dgamdc + \
                    D1H12 * dgamdc + D2H12 * dgamcdc - \
                    D1H13 * dgamcdc - D2H13 * dgamcdc - \
                    D1H14 * dgamdc - D2H14 * dgamdc

            grad2 = D1H21 * dgamdc + D2H21 * dgamcdc + \
                    D1H22 * dgamcdc + D2H22 * dgamdc - \
                    D1H23 * dgamcdc - D2H23 * dgamcdc - \
                    D1H24 * dgamdc - D2H24 * dgamdc

            # gradient of H wrt D
            D_grad1 = D1H11 * dgamcdd + D2H11 * dgamdd + \
                      D1H12 * dgamdd + D2H12 * dgamcdd - \
                      D1H13 * dgamcdd - D2H13 * dgamcdd - \
                      D1H14 * dgamdd - D2H14 * dgamdd

            D_grad2 = D1H21 * dgamdd + D2H21 * dgamcdd + \
                      D1H22 * dgamcdd + D2H22 * dgamdd - \
                      D1H23 * dgamcdd - D2H23 * dgamcdd - \
                      D1H24 * dgamdd - D2H24 * dgamdd            

            

            H1 = H11 + H12 - H13 - H14
            # transpose the axis for H1
            H1 = H1.transpose((1, 0, 3, 2, 4))
            H2 = H21 + H22 - H23 - H24

            H = H1 + H2
            grad1 = grad1.transpose((1, 0, 3, 2, 4, 5))
            D_grad = D_grad1.transpose((1, 0, 3, 2, 4, 5)) + \
                     D_grad2            

            # [..., p, q, i, r] := d K_pq / d S_ir
            K_S_grad = np.zeros((N1, N2, Q, Q, Q, R), dtype=np.complex128)
            for p in range(Q):
                for q in range(Q):
                    K_S_grad[..., p, q, p, :] += (S[q, :] * np.sqrt(np.pi) * length_scale) * \
                                                 (H1[..., p, q, :] + H2[..., p, q, :])
                    K_S_grad[..., p, q, q, :] += (S[p, :] * np.sqrt(np.pi) * length_scale) * \
                                                 (H1[..., p, q, :] + H2[..., p, q, :])
            K_S_grad /= 8 * wprod[None, None, ..., None, None]
            K_S_grad = np.real(K_S_grad)            

            # [N1, N2, K, K, R, K]
            # Utility expression
            expr = Sprod * np.sqrt(np.pi) * length_scale / (8 * wprod[..., None])            
            expr_C_grad = np.zeros((Q, Q, Q), dtype=np.complex128)
            expr_D_grad = np.zeros((Q, Q, Q), dtype=np.complex128)
            for k in range(Q):
                expr_C_grad[k, :, k] -= dwdc[k] / (w[k]**2*w)
                expr_C_grad[:, k, k] -= dwdc[k] / (w[k]**2*w)
                expr_D_grad[k, :, k] -= dwdd[k] / (w[k]**2*w)
                expr_D_grad[:, k, k] -= dwdd[k] / (w[k]**2*w)

            # add an axis for the length scale
            expr_C_grad = expr_C_grad[:, :, None, :]
            expr_C_grad = expr_C_grad * (Sprod * np.sqrt(np.pi) * length_scale / 8)[..., None]

            expr_D_grad = expr_D_grad[:, :, None, :]
            expr_D_grad = expr_D_grad * (Sprod * np.sqrt(np.pi) * length_scale / 8)[..., None]


            K_C_grad = expr[None, None, ..., None] * (grad1 + grad2) + \
                       expr_C_grad[None, None, ...] * (H1 + H2)[..., None]

            K_D_grad = expr[None, None, ..., None] * (D_grad) + \
                       expr_D_grad[None, None, ...] * (H1 + H2)[..., None]

            K = expr[None, None, ...] * (H1 + H2)

            # sum over length scales
            K = K.sum(-1)
            K_C_grad = K_C_grad.sum(-2)
            K_D_grad = K_D_grad.sum(-2)

            # flatten everything out
            K = K.transpose((2, 0, 3, 1)).reshape((Q*N1, Q*N2))

            KdC = K_C_grad.transpose((2, 0, 3, 1, 4)).reshape((Q*N1, Q*N2, Q))
            KdD = K_D_grad.transpose((2, 0, 3, 1, 4)).reshape((Q*N1, Q*N2, Q))            
            KdS = K_S_grad.transpose((2, 0, 3, 1, 4, 5)).reshape((Q*N1, Q*N2, Q, R))

            # further flatten KdS to agree vec(S) = S.T.ravel()
            # [ S_11, ..., S_Q1, ..., ]
            KdS = np.stack([KdS[..., k, r] for r in range(R) for k in range(Q)],
                           axis=-1)

            K_grad = np.dstack((KdC, KdD, KdS))
            return np.real(K), np.real(K_grad)

        else:
            H1 = Aicht2(t1, t2, gamc, gam, length_scale) + \
                 Aicht2(t1, t2, gam, gamc, length_scale) - \
                 Aicht2(t1, t2, gamc, gamc, length_scale) - \
                 Aicht2(t1, t2, gam, gam, length_scale)
            H2 = Aicht2(t2, t1, gam, gamc, length_scale) + \
                 Aicht2(t2, t1, gamc, gam, length_scale) - \
                 Aicht2(t2, t1, gamc, gamc, length_scale) - \
                 Aicht2(t2, t1, gam, gam, length_scale)
            K = H1.transpose((1, 0, 3, 2, 4)) + H2

            expr = Sprod * np.sqrt(np.pi) * length_scale / (8 * wprod[..., None])
            K = K * expr[None, None, ...]
            K = K.sum(-1)

            K = np.real(K)
            # flatten K out
            K = K.transpose((2, 0, 3, 1)).reshape((Q*t1.size,
                                                   Q*t2.size))
            return K

class LFMorder2(GaussianProcessRegressor):
    
    
    def fit(self, X, y):

        self.kernel_ = clone(self.kernel)

        if self.normalize_y:
            self._y_train_mean = np.mean(y, axis=0)
            y = y - self._y_train_mean
        else:
            self._y_train_mean = np.zeros(1)

        if np.iterable(self.alpha) \
           and self.alpha.shape[0] != y.shape[0]:
            if self.alpha.shape[0] == 1:
                self.alpha = self.alpha[0]
            else:
                raise ValueError("alpha must be a scalar of an array"
                                 " with same number of entries as y (%d != %d)"
                                 % (self.alpha.shape[0], y.shape[0]))

        self.X_train_ = np.copy(X) if self.copy_X_train else X
        self.y_train_ = np.copy(y) if self.copy_X_train else y

        if self.optimizer is not None and self.kernel_.n_dims > 0:
            # Choose hyperparameters based on maximizing the log-marginal
            # likelihood (potentially starting from several initial values)
            def obj_func(theta, eval_gradient=True):
                try:
                    if eval_gradient:
                        lml, grad = self.log_marginal_likelihood(
                            theta, eval_gradient=True)
                        print(lml)
                        return -lml, -grad
                    else:
                        return -self.log_marginal_likelihood(theta)
                except KeyError:
                    return np.inf, np.zeros(theta.size)

            from scipy.optimize import minimize
            res = minimize(obj_func, x0=self.kernel_.theta, jac=True)

            self.kernel_ = self.kernel.clone_with_theta(res.x)



import os
import json
import numpy as np

class MocapExample:

    @staticmethod
    def load_data(name):

        pathname = os.path.join(os.path.dirname(__file__),
                                'example_data/{}.json'.format(name))

        try:
            with open(pathname, 'r') as f:
                data = json.load(f)

            for key, item in data.items():
                data[key] = np.array(item)

            return data

        except FileNotFoundError as e:
            raise e

